# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Channel objects.

.. currentmodule:: curious.dataclasses.channel
"""
import collections
import enum
import pathlib
import time
import typing as _typing
from math import floor
from os import PathLike
from types import MappingProxyType
from typing import AsyncIterator

import multio
from async_generator import asynccontextmanager

from curious.dataclasses import guild as dt_guild, invite as dt_invite, member as dt_member, \
    message as dt_message, permissions as dt_permissions, role as dt_role, user as dt_user, \
    webhook as dt_webhook
from curious.dataclasses.bases import Dataclass, IDObject
from curious.dataclasses.embed import Embed
from curious.exc import CuriousError, ErrorCode, Forbidden, HTTPException, PermissionsError
from curious.util import AsyncIteratorWrapper, base64ify, deprecated, safe_generator


class ChannelType(enum.IntEnum):
    """
    Returns a mapping from Discord channel type.
    """

    #: Represents a text channel.
    TEXT = 0

    #: Represents a private channel.
    PRIVATE = 1

    #: Represents a voice channel.
    VOICE = 2

    #: Represents a group channel.
    GROUP = 3

    #: Represents a category channel.
    CATEGORY = 4

    def has_messages(self) -> bool:
        """
        :return: If this channel type has messages.
        """
        return self not in [ChannelType.VOICE, ChannelType.CATEGORY]


class HistoryIterator(collections.AsyncIterator):
    """
    An iterator that allows you to automatically fetch messages and async iterate over them.
    
    .. code-block:: python3
    
        it = HistoryIterator(some_channel, bot, max_messages=100)
        
        # usage 1
        async for message in it:
            ...
            
        # usage 2
        await it.fill_messages()
        for message in it.messages:
            ...
            
    Note that usage 2 will only fill chunks of 100 messages at a time.
    """

    def __init__(self, channel: 'Channel',
                 max_messages: int = -1, *,
                 before: int = None, after: int = None):
        """
        :param channel: The :class:`.Channel` to iterate over.
        :param max_messages: The maximum number of messages to return. <= 0 means infinite.
        :param before: The message ID to fetch before.
        :param after: The message ID to fetch after.

        .. versionchanged:: 0.7.0

            Removed the ``client`` parameter.
        """
        self.channel = channel

        #: The current storage of messages.
        self.messages = collections.deque()

        #: The current count of messages iterated over.
        #: This is used to know when to automatically fill new messages.
        self.current_count = 0

        #: The maximum amount of messages to use.
        #: If this is <= 0, an infinite amount of messages are returned.
        self.max_messages = max_messages

        #: The message ID of before to fetch.
        self.before = before
        if isinstance(self.before, IDObject):
            self.before = self.before.id

        #: The message ID of after to fetch.
        self.after = after
        if isinstance(self.after, IDObject):
            self.after = self.after.id

        #: The last message ID that we fetched.
        if self.before:
            self.last_message_id = self.before
        else:
            self.last_message_id = self.after

    async def fill_messages(self) -> None:
        """
        Called to fill the next <n> messages.
        
        This is called automatically by :meth:`.__anext__`, but can be used to fill the messages
        anyway.
        """
        if self.max_messages < 0:
            to_get = 100
        else:
            to_get = self.max_messages - self.current_count

        if to_get <= 0:
            return

        if self.before:
            messages = await self.channel._bot.http.get_message_history(self.channel.id,
                                                                        before=self.last_message_id,
                                                                        limit=to_get)
        else:
            messages = await self.channel._bot.http.get_message_history(self.channel.id,
                                                                        after=self.last_message_id)
            messages = reversed(messages)

        for message in messages:
            self.messages.append(self.channel._bot.state.make_message(message))

    async def __anext__(self) -> 'dt_message.Message':
        self.current_count += 1
        if self.current_count == self.max_messages:
            raise StopAsyncIteration

        if len(self.messages) <= 0:
            await self.fill_messages()

        try:
            message = self.messages.popleft()
        except IndexError:
            # No messages to fill, so self._fill_messages didn't return any
            # This signals the end of iteration.
            raise StopAsyncIteration
        self.last_message_id = message.id

        return message

    def __iter__(self) -> None:
        raise RuntimeError("This is not an iterator - you want to use `async for` instead.")

    def __await__(self) -> None:
        raise RuntimeError("This is not a coroutine - you want to use `async for` instead.")

    async def next(self) -> 'dt_message.Message':
        """
        Gets the next item in history.
        """
        return await self.__anext__()

    async def all(self) -> '_typing.List[dt_message.Message]':
        """
        Gets a flattened list of items from the history.
        """
        items = []

        async for item in self:
            items.append(item)

        return items


class ChannelMessageWrapper(object):
    """
    Represents a channel's message container.
    """

    __slots__ = "channel",

    def __init__(self, channel: 'Channel'):
        #: The :class:`.Channel` this container is used for.
        self.channel = channel

    def __iter__(self) -> None:
        raise RuntimeError("Use `async for`")

    def __aiter__(self) -> HistoryIterator:
        return self.history.__aiter__()

    @property
    def history(self) -> HistoryIterator:
        """
        :return: A :class:`.HistoryIterator` that can be used to iterate over the channel history.
        """
        return self.get_history(before=self.channel._last_message_id, limit=-1)

    def get_history(self, before: int = None,
                    after: int = None,
                    limit: int = 100) -> HistoryIterator:
        """
        Gets history for this channel.

        This is *not* an async function - it returns a :class:`HistoryIterator` which can be async
        iterated over to get message history.

        .. code-block:: python3

            async for message in channel.get_history(limit=1000):
                print(message.content, "by", message.author.user.name)

        :param limit: The maximum number of messages to get.
        :param before: The snowflake ID to get messages before.
        :param after: The snowflake ID to get messages after.
        """
        if self.channel.guild:
            if not self.channel.permissions(self.channel.guild.me).read_message_history:
                raise PermissionsError("read_message_history")

        return HistoryIterator(self.channel, before=before, after=after, max_messages=limit)

    async def send(self, content: str = None, *,
                   tts: bool = False, embed: 'Embed' = None) -> 'dt_message.Message':
        """
        Sends a message to this channel.

        This requires SEND_MESSAGES permission in the channel.
        If the content is not a string, it will be automatically stringified.

        .. code:: python

            await channel.send("Hello, world!")

        :param content: The content of the message to send.
        :param tts: Should this message be text to speech?
        :param embed: An embed object to send with this message.
        :return: A new :class:`.Message` object.
        """
        if not self.channel.type.has_messages():
            raise CuriousError("Cannot send messages to a voice channel")

        if self.channel.guild:
            if not self.channel.permissions(self.channel.guild.me).send_messages:
                raise PermissionsError("send_messages")

        if not isinstance(content, str) and content is not None:
            content = str(content)

        # check for empty messages
        if not content:
            if not embed:
                raise CuriousError("Cannot send an empty message")

            if self.channel.guild and not \
                    self.channel.permissions(self.channel.guild.me).embed_links:
                raise PermissionsError("embed_links")
        else:
            if content and len(content) > 2000:
                raise CuriousError("Content must be less than 2000 characters")

        if embed is not None:
            embed = embed.to_dict()

        data = await self.channel._bot.http.send_message(self.channel.id, content,
                                                         tts=tts, embed=embed)
        obb = self.channel._bot.state.make_message(data, cache=True)

        return obb

    async def upload(self, fp: '_typing.Union[bytes, str, PathLike, _typing.IO]',
                     *,
                     filename: str = None,
                     message_content: '_typing.Optional[str]' = None) -> 'dt_message.Message':
        """
        Uploads a message to this channel.

        This requires SEND_MESSAGES and ATTACH_FILES permission in the channel.

        .. code-block:: python3

            with open("/tmp/emilia_best_girl.jpg", 'rb') as f:
                await channel.messages.upload(f, "my_waifu.jpg")

        :param fp: Variable.

            - If passed a string or a :class:`os.PathLike`, will open and read the file and
            upload it.
            - If passed bytes, will use the bytes as the file content.
            - If passed a file-like, will read and use the content to upload.

        :param filename: The filename for the file uploaded. If a path-like or str is passed, \
            will use the filename from that if this is not specified.
        :param message_content: Optional: Any extra content to be sent with the message.
        :return: The new :class:`.Message` created.
        """
        if not self.channel.type.has_messages():
            raise CuriousError("Cannot send messages to a voice channel")

        if self.channel.guild:
            if not self.channel.permissions(self.channel.guild.me).send_messages:
                raise PermissionsError("send_messages")

            if not self.channel.permissions(self.channel.guild.me).attach_files:
                raise PermissionsError("attach_files")

        if isinstance(fp, bytes):
            file_content = fp
        elif isinstance(fp, pathlib.Path):
            if filename is None:
                filename = fp.parts[-1]

            file_content = fp.read_bytes()
        elif isinstance(fp, (str, PathLike)):
            path = pathlib.Path(fp)
            if filename is None:
                filename = path.parts[-1]

            file_content = path.read_bytes()
        elif isinstance(fp, _typing.IO) or hasattr(fp, "read"):
            file_content = fp.read()

            if isinstance(file_content, str):
                file_content = file_content.encode("utf-8")
        else:
            raise ValueError("Got unknown type for upload")

        if filename is None:
            filename = "unknown.bin"

        data = await self.channel._bot.http.send_file(self.channel.id, file_content,
                                                      filename=filename, content=message_content)
        obb = self.channel._bot.state.make_message(data, cache=False)
        return obb

    async def bulk_delete(self, messages: '_typing.List[dt_message.Message]') -> int:
        """
        Deletes messages from a channel.
        This is the low-level delete function - for the high-level function, see
        :meth:`.Channel.messages.purge()`.

        Example for deleting all the last 100 messages:

        .. code:: python

            history = channel.messages.get_history(limit=100)
            messages = []

            async for message in history:
                messages.append(message)

            await channel.messages.bulk_delete(messages)

        :param messages: A list of :class:`.Message` objects to delete.
        :return: The number of messages deleted.
        """
        if self.channel.guild:
            if not self.channel.permissions(self.channel.guild.me).manage_messages:
                raise PermissionsError("manage_messages")

        minimum_allowed = floor((time.time() - 14 * 24 * 60 * 60) * 1000.0 - 1420070400000) << 22
        ids = []
        for message in messages:
            if message.id < minimum_allowed:
                msg = f"Cannot delete message id {message.id} older than {minimum_allowed}"
                raise CuriousError(msg)

            ids.append(message.id)

        await self.channel._bot.http.delete_multiple_messages(self.channel.id, ids)

        return len(ids)

    async def purge(self, limit: int = 100, *,
                    author: 'dt_member.Member' = None,
                    content: str = None,
                    predicate: '_typing.Callable[[dt_message.Message], bool]' = None,
                    fallback_from_bulk: bool = False):
        """
        Purges messages from a channel.
        This will attempt to use ``bulk-delete`` if possible, but otherwise will use the normal
        delete endpoint (which can get ratelimited severely!) if ``fallback_from_bulk`` is True.

        Example for deleting all messages owned by the bot:

        .. code-block:: python3

            me = channel.guild.me
            await channel.messages.purge(limit=100, author=me)

        Custom check functions can also be applied which specify any extra checks. They take one
        argument (the Message object) and return a boolean (True or False) determining if the
        message should be deleted.

        For example, to delete all messages with the letter ``i`` in them:

        .. code-block:: python3

            await channel.messages.purge(limit=100,
                                         predicate=lambda message: 'i' in message.content)

        :param limit: The maximum amount of messages to delete. -1 for unbounded size.
        :param author: Only delete messages made by this author.
        :param content: Only delete messages that exactly match this content.
        :param predicate: A callable that determines if a message should be deleted.
        :param fallback_from_bulk: If this is True, messages will be regular deleted if they \
            cannot be bulk deleted.
        :return: The number of messages deleted.
        """
        if self.channel.guild:
            if not self.channel.permissions(self.channel.guild.me).manage_messages \
                    and not fallback_from_bulk:
                raise PermissionsError("manage_messages")

        checks = []
        if author:
            checks.append(lambda m: m.author == author)

        if content:
            checks.append(lambda m: m.content == content)

        if predicate:
            checks.append(predicate)

        to_delete = []
        history = self.get_history(limit=limit)

        async for message in history:
            if all(check(message) for check in checks):
                to_delete.append(message)

        can_bulk_delete = True

        # Split into chunks of 100.
        message_chunks = [to_delete[i:i + 100] for i in range(0, len(to_delete), 100)]
        minimum_allowed = floor((time.time() - 14 * 24 * 60 * 60) * 1000.0 - 1420070400000) << 22
        for chunk in message_chunks:
            message_ids = []
            for message in chunk:
                if message.id < minimum_allowed:
                    msg = f"Cannot delete message id {message.id} older than {minimum_allowed}"
                    raise CuriousError(msg)

                message_ids.append(message.id)

            # First, try and bulk delete all the messages.
            if can_bulk_delete:
                try:
                    await self.channel._bot.http.delete_multiple_messages(self.channel.id,
                                                                          message_ids)
                except Forbidden:
                    # We might not have MANAGE_MESSAGES.
                    # Check if we should fallback on normal delete.
                    can_bulk_delete = False
                    if not fallback_from_bulk:
                        # Don't bother, actually.
                        raise

            # This is an `if not` instead of an `else` because `can_bulk_delete` might've changed.
            if not can_bulk_delete:
                # Instead, just delete() the message.
                for message in chunk:
                    await message.delete()

        return len(to_delete)

    async def get(self, message_id: int) -> 'dt_message.Message':
        """
        Gets a single message from this channel.

        .. versionchanged:: 0.7.0

            Errors raised are now consistent across bots and userbots.

        :param message_id: The message ID to retrieve.
        :return: A new :class:`.Message` object.
        :raises CuriousError: If the message could not be found.
        """
        if self.channel.guild:
            if not self.channel.permissions(self.channel.guild.me).read_message_history:
                raise PermissionsError("read_message_history")

        cached_message = self.channel._bot.state.find_message(message_id)
        if cached_message is not None:
            return cached_message

        try:
            data = await self.channel._bot.http.get_message(self.channel.id, message_id)
        except HTTPException as e:
            # transform into a CuriousError if it wasn't found
            if e.error_code == ErrorCode.UNKNOWN_MESSAGE:
                raise CuriousError("No message found for this ID") from e

            raise

        msg = self.channel._bot.state.make_message(data)

        return msg


class Channel(Dataclass):
    """
    Represents a channel object.
    """

    def __init__(self, client, **kwargs) -> None:
        super().__init__(kwargs.get("id"), client)

        #: The name of this channel.
        self.name = kwargs.get("name", None)  # type: str

        #: The topic of this channel.
        self.topic = kwargs.get("topic", None)  # type: str

        #: The ID of the guild this is associated with.
        self.guild_id = int(kwargs.get("guild_id", 0)) or None  # type: int

        parent_id = kwargs.get("parent_id")
        if parent_id is not None:
            parent_id = int(parent_id)

        #: The parent ID of this channel.
        self.parent_id = parent_id  # type: int

        #: The :class:`.ChannelType` of channel this channel is.
        self.type = ChannelType(kwargs.get("type", 0))  # type: ChannelType

        #: The :class:`.ChannelMessageWrapper` for this channel.
        self._messages = None  # type: ChannelMessageWrapper

        #: If this channel is NSFW.
        self.nsfw = kwargs.get("nsfw", False)  # type: bool

        #: If private, the mapping of :class:`.User` that are in this channel.
        self._recipients = {}  # type: _typing.Dict[int, dt_user.User]

        if self.private:
            for recipient in kwargs.get("recipients", []):
                u = self._bot.state.make_user(recipient)
                self._recipients[u.id] = u

            if self.type == ChannelType.GROUP:
                # append the current user
                self._recipients[self._bot.user.id] = self._bot.user

        #: The position of this channel.
        self.position = kwargs.get("position", 0)  # type: int

        #: The last message ID of this channel.
        #: Used for history.
        self._last_message_id = None  # type: int

        _last_message_id = kwargs.get("last_message_id", 0)
        if _last_message_id:
            self._last_message_id = int(_last_message_id)
        else:
            self._last_message_id = None

        # group channel stuff
        #: The owner ID of the channel.
        #: This is None for non-group channels.
        self.owner_id = int(kwargs.get("owner_id", 0)) or None  # type: int

        #: The icon hash of the channel.
        self.icon_hash = kwargs.get("icon", None)  # type: str

        #: The internal overwrites for this channel.
        self._overwrites = {}  # type: _typing.Dict[int, dt_permissions.Overwrite]

    def __repr__(self) -> str:
        return f"<Channel id={self.id} name={self.name} type={self.type.name} " \
               f"guild_id={self.guild_id}>"

    __str__ = __repr__

    def _update_overwrites(self, overwrites: _typing.List[dict]):
        """
        Updates the overwrites for this channel.

        :param overwrites: A list of overwrite dicts.
        """
        if not self.guild_id:
            raise CuriousError("A channel without a guild cannot have overwrites")

        self._overwrites = {}

        for overwrite in overwrites:
            id_ = int(overwrite["id"])
            type_ = overwrite["type"]

            if type_ == "member":
                obb = self.guild._members.get(id_)
            else:
                obb = self.guild._members.get(id_)

            self._overwrites[id] = dt_permissions.Overwrite(allow=overwrite["allow"],
                                                            deny=overwrite["deny"],
                                                            obb=obb, channel_id=self.id)

    @property
    def guild(self) -> '_typing.Union[dt_guild.Guild, None]':
        """
        :return: The :class:`.Guild` associated with this Channel.
        """
        try:
            return self._bot.guilds[self.guild_id]
        except KeyError:
            return None

    @property
    def private(self) -> bool:
        """
        :return: If this channel is a private channel (i.e has no guild.)
        """
        return self.guild_id is None

    @property
    def recipients(self) -> '_typing.Mapping[int, dt_user.User]':
        """
        :return: A mapping of int -> :class:`.User` for the recipients of this private chat.
        """
        return MappingProxyType(self._recipients)

    @property
    def user(self) -> '_typing.Union[dt_user.User, None]':
        """
        :return: If this channel is a private channel, the :class:`.User` of the other user.
        """
        if self.type != ChannelType.PRIVATE:
            return None

        try:
            return next(iter(self.recipients.values()))
        except StopIteration:
            return None

    @property
    def owner(self) -> '_typing.Union[dt_user.User, None]':
        """
        :return: If this channel is a group channel, the owner of the channel.
        """
        if not self.owner_id:
            return None

        try:
            return self._bot.state._users[self.owner_id]
        except KeyError:
            return None

    @property
    def parent(self) -> '_typing.Union[Channel, None]':
        """
        :return: If this channel has a parent, the parent category of this channel.
        """
        try:
            return self.guild.channels[self.parent_id]
        except (KeyError, AttributeError):
            return None

    @property
    def children(self) -> '_typing.List[Channel]':
        """
        :return: A list of :class:`.Channel` children this channel has, if any.
        """
        if not self.guild:
            return []

        channels = [channel
                    for channel in self.guild.channels.values()
                    if channel.parent_id == self.id]
        return channels

    def get_by_name(self, name: str) -> '_typing.Union[Channel, None]':
        """
        Gets a channel by name in this channel's children.

        :param name: The name of the channel to get.
        :return: A :class:`.Channel` if the channel was find
        """
        return next(filter(lambda channel: channel.name == name, self.children), None)

    @property
    def messages(self) -> 'ChannelMessageWrapper':
        """
        :return: The :class:`.ChannelMessageWrapper` for this channel, if applicable.
        """
        if not self.type.has_messages():
            raise CuriousError("This channel does not have messages")

        if self._messages is None:
            self._messages = ChannelMessageWrapper(self)
        return self._messages

    @property
    @deprecated(since="0.7.0", see_instead="Channel.messages", removal="0.9.0")
    def history(self) -> HistoryIterator:
        """
        :return: A :class:`.HistoryIterator` that can be used to iterate over the channel history.
        """
        return self.messages.history

    @property
    def pins(self) -> '_typing.AsyncIterator[dt_message.Message]':
        """
        :return: A :class:`.AsyncIteratorWrapper` that can be used to iterate over the pins.
        """
        return AsyncIteratorWrapper(self.get_pins)

    @property
    def icon_url(self) -> _typing.Union[str, None]:
        """
        :return: The icon URL for this channel if it is a group DM. 
        """
        return "https://cdn.discordapp.com/channel-icons/{}/{}.webp" \
            .format(self.id, self.icon_hash)

    @property
    def voice_members(self) -> '_typing.List[dt_member.Member]':
        """
        :return: A list of members that are in this voice channel.
        """
        if self.type != ChannelType.VOICE:
            raise CuriousError("No members for channels that aren't voice channels")

        return list(
            filter(lambda member: member.voice.channel == self, self.guild.members.values())
        )

    @property
    def overwrites(self) -> '_typing.Mapping[int, dt_permissions.Overwrite]':
        """
        :return: A mapping of target_id -> :class:`.Overwrite` for this channel.
        """
        return MappingProxyType(self._overwrites)

    def permissions(self, obb: '_typing.Union[dt_member.Member, dt_role.Role]') -> \
            'dt_permissions.Overwrite':
        """
        Gets the permission overwrites for the specified object.
        """
        overwrite = self._overwrites.get(obb.id)
        if not overwrite:
            overwrite = dt_permissions.Overwrite(0, 0, obb)
            overwrite.channel_id = self.id

        return overwrite

    @property
    def me_permissions(self) -> 'dt_permissions.Overwrite':
        """
        :return: The overwrite permissions for the current member.
        """
        return self.permissions(self.guild.me)

    def _copy(self):
        obb = object.__new__(self.__class__)
        obb.name = self.name
        obb.type = self.type
        obb.guild_id = self.guild_id
        obb.nsfw = self.nsfw
        obb._recipients = self._recipients
        obb.icon_hash = self.icon_hash
        obb.owner_id = self.owner_id
        obb.topic = self.topic
        obb.position = self.position
        obb._bot = self._bot
        obb.parent_id = self.parent_id
        return obb

    @deprecated(since="0.7.0", see_instead="Channel.messages.get_history", removal="0.9.0")
    def get_history(self, before: int = None,
                    after: int = None,
                    limit: int = 100) -> HistoryIterator:
        """
        Gets history for this channel.

        This is *not* an async function - it returns a :class:`HistoryIterator` which can be async
        iterated over to get message history.

        .. code-block:: python3

            async for message in channel.get_history(limit=1000):
                print(message.content, "by", message.author.user.name)

        :param limit: The maximum number of messages to get.
        :param before: The snowflake ID to get messages before.
        :param after: The snowflake ID to get messages after.
        """
        return self.messages.get_history(before=before, after=after, limit=limit)

    async def get_pins(self) -> '_typing.List[dt_message.Message]':
        """
        Gets the pins for a channel.

        :return: A list of :class:`.Message` objects.
        """
        msg_data = await self._bot.http.get_pins(self.id)

        messages = []
        for message in msg_data:
            messages.append(self._bot.state.make_message(message))

        return messages

    @property
    def webhooks(self) -> 'AsyncIterator[dt_webhook.Webhook]':
        """
        :return: A :class:`.AsyncIteratorWrapper` for the :class:`.Webhook` objects in this \
            channel.
        """
        return AsyncIteratorWrapper(self.get_webhooks)

    async def get_webhooks(self) -> '_typing.List[dt_webhook.Webhook]':
        """
        Gets the webhooks for this channel.

        :return: A list of :class:`.Webhook` objects for the channel.
        """
        webhooks = await self._bot.http.get_webhooks_for_channel(self.id)
        obbs = []

        for webhook in webhooks:
            obbs.append(self._bot.state.make_webhook(webhook))

        return obbs

    @deprecated(since="0.7.0", see_instead="Channel.messages.get", removal="0.9.0")
    async def get_message(self, message_id: int) -> 'dt_message.Message':
        """
        Gets a single message from this channel.

        :param message_id: The message ID to retrieve.
        :return: A new :class:`.Message` object.
        """
        return await self.messages.get(message_id)

    async def create_webhook(self, *, name: str = None,
                             avatar: bytes = None) -> 'dt_webhook.Webhook':
        """
        Create a webhook in this channel.

        :param name: The name of the new webhook.
        :param avatar: The bytes content of the new webhook.
        :return: A :class:`.Webhook` that represents the webhook created.
        """
        if not self.permissions(self.guild.me).manage_webhooks:
            raise PermissionsError("manage_webhooks")

        if avatar is not None:
            avatar = base64ify(avatar)

        data = await self._bot.http.create_webhook(self.id, name=name, avatar=avatar)
        webook = self._bot.state.make_webhook(data)

        return webook

    async def edit_webhook(self, webhook: 'dt_webhook.Webhook', *,
                           name: str = None, avatar: bytes = None) -> 'dt_webhook.Webhook':
        """
        Edits a webhook.

        :param webhook: The :class:`.Webhook` to edit.
        :param name: The new name for the webhook.
        :param avatar: The new bytes for the avatar.
        :return: The modified :class:`.Webhook`. object.
        """
        if avatar is not None:
            avatar = base64ify(avatar)

        if webhook.token is not None:
            # Edit it unconditionally.
            await self._bot.http.edit_webhook_with_token(webhook.id, webhook.token,
                                                         name=name, avatar=avatar)

        if not self.permissions(self.guild.me).manage_webhooks:
            raise PermissionsError("manage_webhooks")

        data = await self._bot.http.edit_webhook(webhook.id,
                                                 name=name, avatar=avatar)
        webhook.default_name = data.get("name")
        webhook._default_avatar = data.get("avatar")

        webhook.user.username = data.get("name")
        webhook.user.avatar_hash = data.get("avatar")

        return webhook

    async def delete_webhook(self, webhook: 'dt_webhook.Webhook') -> 'dt_webhook.Webhook':
        """
        Deletes a webhook.

        You must have MANAGE_WEBHOOKS to delete this webhook.

        :param webhook: The :class:`.Webhook` to delete.
        """
        if webhook.token is not None:
            # Delete it unconditionally.
            await self._bot.http.delete_webhook_with_token(webhook.id, webhook.token)
            return webhook

        if not self.permissions(self.guild.me).manage_webhooks:
            raise PermissionsError("manage_webhooks")

        await self._bot.http.delete_webhook(webhook.id)
        return webhook

    async def create_invite(self, **kwargs) -> 'dt_invite.Invite':
        """
        Creates an invite in this channel.

        :param max_age: The maximum age of the invite.
        :param max_uses: The maximum uses of the invite.
        :param temporary: Is this invite temporary?
        :param unique: Is this invite unique?
        """
        if not self.guild:
            raise PermissionsError("create_instant_invite")

        if not self.permissions(self.guild.me).create_instant_invite:
            raise PermissionsError("create_instant_invite")

        inv = await self._bot.http.create_invite(self.id, **kwargs)
        invite = dt_invite.Invite(self._bot, **inv)

        return invite

    @deprecated(since="0.7.0", see_instead="Channel.messages.delete_messages",
                removal="0.9.0")
    async def delete_messages(self, messages: '_typing.List[dt_message.Message]') -> int:
        """
        Deletes messages from a channel.
        This is the low-level delete function - for the high-level function, see 
        :meth:`.Channel.purge()`.

        Example for deleting all the last 100 messages:

        .. code:: python

            history = channel.get_history(limit=100)
            messages = []

            async for message in history:
                messages.append(message)

            await channel.delete_messages(messages)

        :param messages: A list of :class:`.Message` objects to delete.
        :return: The number of messages deleted.
        """
        return await self.messages.bulk_delete(messages)

    @deprecated(since="0.7.0", see_instead="Channel.messages.purge", removal="0.9.0")
    async def purge(self, limit: int = 100, *,
                    author: 'dt_member.Member' = None,
                    content: str = None,
                    predicate: '_typing.Callable[[dt_message.Message], bool]' = None,
                    fallback_from_bulk: bool = False):
        """
        Purges messages from a channel.
        This will attempt to use ``bulk-delete`` if possible, but otherwise will use the normal
        delete endpoint (which can get ratelimited severely!) if ``fallback_from_bulk`` is True.

        Example for deleting all messages owned by the bot:

        .. code-block:: python3

            me = channel.guild.me
            await channel.purge(limit=100, author=me)

        Custom check functions can also be applied which specify any extra checks. They take one
        argument (the Message object) and return a boolean (True or False) determining if the
        message should be deleted.

        For example, to delete all messages with the letter ``i`` in them:

        .. code-block:: python3

            await channel.purge(limit=100, predicate=lambda message: 'i' in message.content)

        :param limit: The maximum amount of messages to delete. -1 for unbounded size.
        :param author: Only delete messages made by this author.
        :param content: Only delete messages that exactly match this content.
        :param predicate: A callable that determines if a message should be deleted.
        :param fallback_from_bulk: If this is True, messages will be regular deleted if they \
            cannot be bulk deleted.
        :return: The number of messages deleted.
        """
        return await self.messages.purge(limit=limit, author=author, content=content,
                                         predicate=predicate, fallback_from_bulk=fallback_from_bulk)

    async def send_typing(self) -> None:
        """
        Starts typing in the channel for 5 seconds.
        """
        if not self.type.has_messages():
            raise CuriousError("Cannot send messages to this channel")

        if self.guild:
            if not self.permissions(self.guild.me).send_messages:
                raise PermissionsError("send_message")

        await self._bot.http.send_typing(self.id)

    @property
    @asynccontextmanager
    @safe_generator
    async def typing(self) -> _typing.AsyncContextManager[None]:
        """
        :return: A context manager that sends typing repeatedly.

        Usage:

        .. code-block:: python3

            async with channel.typing:
                res = await do_long_action()

            await channel.messages.send("Long action:", res)
        """
        running = multio.Event()

        async def runner():
            await self.send_typing()
            while True:
                try:
                    async with multio.timeout_after(5):
                        await running.wait()
                except multio.asynclib.TaskTimeout:
                    await self.send_typing()
                else:
                    return

        async with multio.asynclib.task_manager() as tg:
            await multio.asynclib.spawn(tg, runner)
            try:
                yield
            finally:
                await multio.asynclib.cancel_task_group(tg)

    @deprecated(since="0.7.0", see_instead="Channel.messages.send", removal="0.10.0")
    async def send(self, content: str = None, *,
                   tts: bool = False, embed: Embed = None) -> 'dt_message.Message':
        """
        Sends a message to this channel.

        This requires SEND_MESSAGES permission in the channel.
        If the content is not a string, it will be automatically stringified.

        .. code:: python

            await channel.send("Hello, world!")

        :param content: The content of the message to send.
        :param tts: Should this message be text to speech?
        :param embed: An embed object to send with this message.
        :return: A new :class:`.Message` object.
        """
        return await self.messages.send(content, tts=tts, embed=embed)

    @deprecated(since="0.7.0", see_instead="Channel.messages.upload", removal="0.10.0")
    async def send_file(self, file_content: bytes, filename: str,
                        *, message_content: _typing.Optional[str] = None) -> 'dt_message.Message':
        """
        Uploads a message to this channel.

        This requires SEND_MESSAGES and ATTACH_FILES permission in the channel.

        .. code:: python

            with open("/tmp/emilia_best_girl.jpg", 'rb') as f:
                await channel.send_file(f.read(), "my_waifu.jpg")

        :param file_content: The bytes-like file content to upload.
            This **cannot** be a file-like object.
            
        :param filename: The filename of the file.
        :param message_content: Optional: Any extra content to be sent with the message.
        :return: The new :class:`.Message` created.
        """
        return await self.messages.upload(file_content, filename, message_content=message_content)

    @deprecated(since="0.7.0", see_instead="Channel.messages.upload", removal="0.10.0")
    async def upload_file(self, filename: str, *,
                          message_content: str = None) -> 'dt_message.Message':
        """
        A higher level interface to ``send_file``.

        This allows you to specify one of the following to upload:

            - A filename (str)
            - A file-like object
            - A path-like object

        This will open the file, read it in binary, and upload it to the channel.

        :param filename: The file to send, in the formats specified above.
        :param message_content: Any extra content to be sent with the message.
        :return: The new :class:`.Message` created.
        """
        return await self.messages.upload(fp=filename, filename=filename,
                                          message_content=message_content)

    async def change_overwrite(self, overwrite: 'dt_permissions.Overwrite'):
        """
        Changes an overwrite for this channel.

        This overwrite must be an instance of :class:`.Overwrite`.

        :param overwrite: The specific overwrite to use.
            If this is None, the overwrite will be deleted.
        """
        if not self.guild:
            raise PermissionsError("manage_roles")

        if not self.permissions(self.guild.me).manage_roles:
            raise PermissionsError("manage_roles")

        target = overwrite.target

        if isinstance(target, dt_member.Member):
            type_ = "member"
        else:
            type_ = "role"

        if overwrite is None:
            # Delete the overwrite instead.
            coro = self._bot.http.remove_overwrite(channel_id=self.id, target_id=target.id)

            async def _listener(before, after):
                if after.id != self.id:
                    return False

                # probably right /shrug
                return True
        else:
            coro = self._bot.http.edit_overwrite(self.id, target.id, type_,
                                                 allow=overwrite.allow.bitfield,
                                                 deny=overwrite.deny.bitfield)

            async def _listener(before, after):
                return after.id == self.id

        async with self._bot.events.wait_for_manager("channel_update", _listener):
            await coro

        return self

    async def edit(self, **kwargs) -> 'Channel':
        """
        Edits this channel.
        """
        if self.guild is None:
            raise CuriousError("Can only edit guild channels")

        if not self.permissions(self.guild.me).manage_channels:
            raise PermissionsError("manage_channels")

        if "type_" in kwargs:
            kwargs["type"] = kwargs["type_"]

        if "type" not in kwargs:
            kwargs["type"] = self.type

        if "parent" in kwargs:
            kwargs["parent_id"] = kwargs["parent"].id

        await self._bot.http.edit_channel(self.id, **kwargs)
        return self

    async def delete(self) -> 'Channel':
        """
        Deletes this channel.
        """
        if not self.permissions(self.guild.me).manage_channels:
            raise PermissionsError("manaqe_channels")

        await self._bot.http.delete_channel(self.id)
        return self

    async def connect(self):
        """
        Connects to voice in this channel.
        """
        if self.type != ChannelType.VOICE:
            raise CuriousError("Cannot connect to a text channel")

        return await self.guild.connect_to_voice(self)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
A reactions-based paginator.
"""
import typing

import multio

from curious.core.event import ListenerExit
from curious.dataclasses.channel import Channel
from curious.dataclasses.embed import Embed
from curious.dataclasses.member import Member
from curious.dataclasses.message import Message
from curious.dataclasses.reaction import Reaction
from curious.dataclasses.user import User


class ReactionsPaginator(object):
    """
    A paginator for a message using reactions.
    """
    BUTTON_BACKWARDS = "◀"
    BUTTON_FORWARD = "▶"
    BUTTON_STOP = "⏹"

    def __init__(self, content: typing.Union[str, typing.List[str], typing.List[Embed]],
                 channel: Channel,
                 respond_to: typing.Union[Member, User], *,
                 break_at: int = 2000, title: str = None):
        """
        :param content: The content to page through.
        :param channel: The channel to send the content to.
        :param respond_to: The member to respond
        :param break_at: The number of characters to break the message up into.
        :param title: The title to put above the embed.
        """
        self._content = content
        self.channel = channel
        self.respond_to = respond_to
        self.title = title

        self.bot = self.channel._bot  # hacky af

        # chunk the message up
        if isinstance(content, list):
            self._message_chunks = content
        else:
            self._message_chunks = [self._content[i:i + break_at] for i in
                                    range(0, len(self._content), break_at)]

        #: The current page this paginator is on.
        self.page = 0

        #: The message object that is being edited.
        self._message = None  # type: Message
        self._running = False
        self._reaction_queue = multio.Queue()

    @classmethod
    async def paginate_response(cls, content: str,
                                responding_to: Message, *args, **kwargs) -> 'ReactionsPaginator':
        """
        Paginates a response to a message.

        :param content: The content to paginate.
        :param responding_to: The message object that you are responding to.
        """
        obb = cls(content, responding_to.channel, responding_to.author, *args, **kwargs)
        await obb.paginate()
        return obb

    async def send_current_page(self) -> None:
        """
        Sends the current page to the channel.
        """
        chunk = self._message_chunks[self.page]
        if isinstance(chunk, Embed):
            embed = chunk
        else:
            embed = Embed(description=self._message_chunks[self.page])
            embed.set_footer(text="Page {}/{}".format(self.page + 1, len(self._message_chunks)))

        if self._message is None:
            self._message = await self.channel.messages.send(
                content=self.title,
                embed=embed
            )
        else:
            await self._message.edit(
                new_content=self.title,
                embed=embed
            )

    async def _add_initial_reactions(self):
        """
        Adds the initial reactions to this message.
        """
        await self._message.react(self.BUTTON_BACKWARDS)
        await self._message.react(self.BUTTON_FORWARD)
        await self._message.react(self.BUTTON_STOP)

    async def paginate(self):
        """
        Starts paginating this message.

        This will continuously listen for reactions on this message until the STOP button is
        pressed.
        """
        self._running = True

        async def consume_reaction(ctx, message: Message, author: Member, reaction: Reaction):
            """
            Consumes reaction events and places them on a queue.
            """
            if message.id != self._message.id:
                return

            if author.id != self.respond_to.id:
                return

            if self._running:
                await self._reaction_queue.put(reaction)
            else:
                raise ListenerExit

        # spawn the consumer task first
        self.bot.events.add_temporary_listener("message_reaction_add", consume_reaction)

        # send the stuff we want
        await self.send_current_page()
        await self._add_initial_reactions()

        try:
            while True:
                async with multio.asynclib.timeout_after(120):
                    reaction = await self._reaction_queue.get()

                if reaction.emoji == self.BUTTON_FORWARD:
                    if self.page < len(self._message_chunks) - 1:
                        self.page += 1
                    else:
                        self.page = 0
                    await self.send_current_page()

                if reaction.emoji == self.BUTTON_BACKWARDS:
                    if self.page > 0:
                        self.page -= 1
                    else:
                        self.page = len(self._message_chunks) - 1

                    await self.send_current_page()

                if reaction.emoji == self.BUTTON_STOP:
                    # remove all reactions were done here
                    break

                await self._message.unreact(reaction.emoji, self.respond_to)
        except multio.asynclib.TaskTimeout:
            # eat timeouts but nothing else
            pass

        self._running = False
        # we've broken out of the loop, so remove reactions and cancel the listener
        await self._message.remove_all_reactions()
        self.bot.events.remove_listener_early("message_reaction_add", consume_reaction)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
The client for an IPC connection.

.. currentmodule:: curious.ipc.client
"""
import os
import platform
import uuid
from typing import Coroutine

import curio
from curio.io import Socket

from curious.dataclasses.presence import RichPresence
from curious.ipc.packet import IPCOpcode, IPCPacket


def get_ipc_url(slot: int = 0) -> str:
    """
    Gets the IPC URL for Discord.
    """
    if platform.system() == "Linux":
        return f"/run/user/{os.getuid()}/discord-ipc-{slot}"
    elif platform.platform() == "Windows":
        return fr"\\?\pipe\discord-ipc-{slot}"


def get_nonce() -> str:
    """
    Gets a random nonce.
    """
    return str(uuid.uuid4())


class IPCError(Exception):
    """
    Represents an error with the IPC protocol.
    """
    pass


class IPCClient(object):
    """
    Represents an IPC (interprocess communication) client. This connects to the Discord client on
    the IPC socket.

    To use, create a new instance with your app's client ID:

    .. code-block:: python3

        ipc = IPCClient(323578534763298816)

    Make sure to open the client before doing anything with it:

    .. code-block:: python3

        await ipc.open()

    """
    VERSION = 1

    def __init__(self, client_id: int, *, slot: int = 0):
        """
        :param client_id: The client ID to authenticate with.
        """
        self.client_id = client_id

        self._ready = False
        self._sock = None  # type: Socket
        self._ipc_slot = slot

    async def open(self):
        """
        Opens this IPC socket.
        """
        path = get_ipc_url(self._ipc_slot)

        self._sock = await curio.open_unix_connection(path)
        await self._write_handshake()
        next_pack = await self.read_packet()

        if next_pack.opcode == IPCOpcode.CLOSE:
            raise IPCError(next_pack.data)

        if next_pack.event != "READY":
            await self._sock.close()
            raise RuntimeError("Didn't receive a READY event.")
        else:
            self._ready = True

        return self

    # Writer methods
    async def _write_packet(self, packet: IPCPacket):
        """
        Writes an IPC packet.

        :param packet: The :class:`.IPCPacket` to write.
        """
        data = packet.serialize()
        await self._sock.sendall(data)

    def _write_json(self, opcode: IPCOpcode, data: dict) -> Coroutine[None, None, None]:
        """
        Writes JSON to the IPC socket.
        """
        return self._write_packet(IPCPacket(opcode, data))

    def _write_handshake(self):
        """
        Writes an IPC handshake.
        """
        data = {
            "v": IPCClient.VERSION,
            "client_id": str(self.client_id)
        }

        return self._write_json(IPCOpcode.HANDSHAKE, data)

    def _write_rich_presence(self, presence: RichPresence):
        """
        Writes a rich presence packet.
        """
        data = {
            "cmd": "SET_ACTIVITY",
            "args": {
                "pid": os.getpid(),
                "activity": presence._rich_fields
            },
            "nonce": get_nonce()
        }
        return self._write_json(IPCOpcode.FRAME, data)

    # Reader methods
    def read_packet(self) -> Coroutine[None, None, IPCPacket]:
        """
        Reads a packet from the connection.
        """
        return IPCPacket.read_packet(self._sock)

    # Convenience methods
    async def send_rich_presence(self, presence: RichPresence):
        """
        Sends a Rich Presence packet to the Discord IPC.

        :param presence: The :class:`.RichPresence` to use.
        """
        if not self._ready:
            raise RuntimeError("Not ready")

        await self._write_rich_presence(presence)
        response = await self.read_packet()

        if response.event == "ERROR":
            raise IPCError(response.data["message"])



"""
An example bot that uses Curious' commands.
"""

# Required imports!
# You have to import the Client (to create the bot) and the Plugin class (to create a plugin).
from curious.commands import CommandsManager, command, condition
from curious.commands.context import Context
from curious.commands.plugin import Plugin
from curious.core.client import Client
from curious.core.event import EventContext


# Plugins are defined as types that inherit from ``Plugin``.
# You can inherit from any other class, but you MUST inherit from Plugin as well.
class Core(Plugin):
    # To define a command, you decorate a regular function with a `command()` decorator.
    # This will mark the function as a command that can be ran inside Discord.

    # The default name for the command will be the same as the function name. You can override
    # this by passing `name=` as a keyword argument to the decorator.
    @command()
    # All commands take a single parameter by default, `ctx`. This is the context which contains
    # useful attributes such as the channel or the guild the command was received on.
    async def ping(self, ctx: Context):
        """
        Responds to a user with `Pong!`.
        """
        # The docstring above is the default help text for the command.
        # We can get the member by using `ctx.author` parameter.
        member = ctx.author
        # To send a message back to the channel, you must use `ctx.channel.send(...)`.
        # We use `member.mention` here to mention the person that pinged us.
        await ctx.channel.messages.send("{}, pong!".format(member.mention))

    # Commands can also take parameters. They are defined by the function signature of the function.
    @command()
    async def hello(self, ctx: Context, name: str):
        """
        Says hello to somebody.
        """
        # The `name` in our signature allows somebody to run `!hello world` and `name` will be
        # automatically replaced with the string `world`.
        await ctx.channel.messages.send("Hello, {}!".format(name))

    # Commands can also provide aliases, by passing in a list.
    @command(aliases=["cold"])
    async def cool(self, ctx: Context):
        """
        Tells if the user is cool or not.
        """
        await ctx.channel.messages.send("It's a bit chilly.")

    # You can also provide invokation checks - these prevent people from running it unless they
    # meet specific criteria.
    @command()
    @condition(lambda ctx: "e" not in ctx.author.nickname)
    async def a(self, ctx: Context):
        """
        Only users without an `e` in their nickname can run this command.
        """
        await ctx.channel.messages.send("The letter `e` is for nerds!")


# To tie this all together, a new client instance needs to be created, and a manager created.
# The command_prefix argument tells us what prefix should be used to invoke commands.
bot = Client()
manager = CommandsManager(client=bot, command_prefix="!")
manager.register_events()


# Add the Core class as a plugin to the bot, inside the ready event.
@bot.event("ready")
async def ready(ctx: EventContext):
    print("Logged in as", bot.user.name)
    await manager.load_plugin(Core)


# Run the bot with your token.
bot.run('MjYwOTUwODE2NTM2NTI2ODQ5.Cz2mGQ.SKl78a6NT6SBpwYQrIDnR1olPqo')



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Special helpers for events.

.. currentmodule: curious.core.events
"""
import functools
import inspect
import logging
import typing

import multio
from async_generator import asynccontextmanager
from multidict import MultiDict

from curious.core import client as md_client
from curious.core.gateway import GatewayHandler
from curious.util import remove_from_multidict, safe_generator

logger = logging.getLogger("curious.events")


class ListenerExit(Exception):
    """
    Raised when a temporary listener is to be exited.

    .. code-block:: python3

        def listener(ctx, message):
            if message.author.id == message.guild.owner_id:
                raise ListenerExit

    """


@asynccontextmanager
@safe_generator
async def _wait_for_manager(manager, name: str, predicate):
    """
    Helper class for managing a wait_for.
    """
    async with multio.asynclib.task_manager() as tg:
        try:
            partial = functools.partial(manager.wait_for, name, predicate)
            await multio.asynclib.spawn(tg, partial)
            yield
        finally:
            await multio.asynclib.cancel_task_group(tg)


class EventManager(object):
    """
    A manager for events.

    This deals with firing of events and temporary listeners.
    """

    def __init__(self):
        #: The task manager used to spawn events.
        self.task_manager = None

        #: A list of event hooks.
        self.event_hooks = set()

        #: A MultiDict of event listeners.
        self.event_listeners = MultiDict()

        #: A MultiDict of temporary listeners.
        self.temporary_listeners = MultiDict()

    # add or removal functions
    # Events
    def add_event(self, func, name: str = None):
        """
        Add an event to the internal registry of events.

        :param name: The event name to register under.
        :param func: The function to add.
        """
        if not inspect.iscoroutinefunction(func):
            raise TypeError("Event must be an async function")

        if name is None:
            evs = func.events
        else:
            evs = [name]

        for ev_name in evs:
            logger.debug("Registered event `{}` handling `{}`".format(func, ev_name))
            self.event_listeners.add(ev_name, func)

    def remove_event(self, name: str, func):
        """
        Removes a function event.

        :param name: The name the event is registered under.
        :param func: The function to remove.
        """
        self.event_listeners = remove_from_multidict(self.event_listeners, key=name, item=func)

    # listeners
    def add_temporary_listener(self, name: str, listener):
        """
        Adds a new temporary listener.

        To remove the listener, you can raise ListenerExit which will exit it and remove the
        listener from the list.

        :param name: The name of the event to listen to.
        :param listener: The listener function.
        """
        self.temporary_listeners.add(name, listener)

    def remove_listener_early(self, name: str, listener):
        """
        Removes a temporary listener early.

        :param name: The name of the event the listener is registered under.
        :param listener: The listener function.
        """
        self.event_listeners = remove_from_multidict(self.event_listeners, key=name, item=listener)

    def add_event_hook(self, listener):
        """
        Adds an event hook.

        :param listener: The event hook callable to use.
        """
        logger.warning("Adding event hook '%s'", listener)
        self.event_hooks.add(listener)

    def remove_event_hook(self, listener):
        """
        Removes an event hook.
        """
        self.event_hooks.remove(listener)

    # wrapper functions
    async def _safety_wrapper(self, func, *args, **kwargs):
        """
        Ensures a coro's error is caught and doesn't balloon out.
        """
        try:
            await func(*args, **kwargs)
        except Exception as e:
            logger.exception("Unhandled exception in {}!".format(func.__name__), exc_info=True)

    async def _listener_wrapper(self, key: str, func, *args, **kwargs):
        """
        Wraps a listener, ensuring ListenerExit is handled properly.
        """
        try:
            await func(*args, **kwargs)
        except ListenerExit:
            # remove the function
            self.temporary_listeners = remove_from_multidict(self.temporary_listeners, key, func)
        except Exception:
            logger.exception("Unhandled exception in listener {}!".format(func.__name__),
                             exc_info=True)
            self.temporary_listeners = remove_from_multidict(self.temporary_listeners, key, func)

    async def wait_for(self, event_name: str, predicate=None):
        """
        Waits for an event.

        Returning a truthy value from the predicate will cause it to exit and return.

        :param event_name: The name of the event.
        :param predicate: The predicate to use to check for the event.
        """
        p = multio.Promise()
        errored = False

        async def listener(ctx, *args):
            # exit immediately if the predicate is none
            if predicate is None:
                await p.set(args)
                raise ListenerExit

            try:
                res = predicate(*args)
                if inspect.isawaitable(res):
                    res = await res
            except ListenerExit:
                # ???
                await p.set(args)
                raise
            except Exception as e:
                # something bad happened, set exception and exit
                logger.exception("Exception in wait_for predicate!")
                # signal that an error happened
                nonlocal errored
                errored = True
                await p.set(e)
                raise ListenerExit
            else:
                # exit now if result is true
                if res is True:
                    await p.set(args)
                    raise ListenerExit

        self.add_temporary_listener(name=event_name, listener=listener)
        output = await p.wait()
        if errored:
            raise output

        # unwrap tuples, if applicable
        if len(output) == 1:
            return output[0]
        return output

    def wait_for_manager(self, event_name: str, predicate) -> 'typing.AsyncContextManager[None]':
        """
        Returns a context manager that can be used to run some steps whilst waiting for a
        temporary listener.

        .. code-block:: python

            async with client.events.wait_for_manager("member_update", predicate=...):
                await member.nickname.set("Test")

        This probably won't be needed outside of internal library functions.
        """
        return _wait_for_manager(self, event_name, predicate)

    async def spawn(self, cofunc, *args) -> typing.Any:
        """
        Spawns a new async function using our task manager.

        Usage::

            async def myfn(a, b):
                await do_some_operation(a + b)

            await events.spawn(myfn, 1, 2)

        :param cofunc: The async function to spawn.
        :param args: Args to provide to the async function.
        """
        return await multio.asynclib.spawn(self.task_manager, cofunc, *args)

    async def fire_event(self, event_name: str, *args, **kwargs):
        """
        Fires an event.

        :param event_name: The name of the event to fire.
        """
        if "ctx" not in kwargs:
            gateway = kwargs.pop("gateway")
            client = kwargs.pop("client")
            ctx = EventContext(client, gateway.gw_state.shard_id, event_name)
        else:
            ctx = kwargs.pop("ctx")

        # clobber event name
        ctx.event_name = event_name

        # always ensure hooks are ran first
        for hook in self.event_hooks:
            cofunc = functools.partial(hook, ctx, *args, **kwargs)
            await self.spawn(cofunc)

        for handler in self.event_listeners.getall(event_name, []):
            coro = functools.partial(handler, ctx, *args, **kwargs)
            coro.__name__ = handler.__name__
            await self.spawn(self._safety_wrapper, coro)

        for listener in self.temporary_listeners.getall(event_name, []):
            coro = functools.partial(self._listener_wrapper, event_name, listener, ctx,
                                     *args, **kwargs)
            await self.spawn(coro)


def event(name, scan: bool = True):
    """
    Marks a function as an event.

    :param name: The name of the event.
    :param scan: Should this event be handled in scans too?
    """

    def __innr(f):
        if not hasattr(f, "events"):
            f.events = {name}

        f.is_event = True
        f.events.add(name)
        f.scan = scan
        return f

    return __innr


def scan_events(obb) -> typing.Generator[None, typing.Tuple[str, typing.Any], None]:
    """
    Scans an object for any items marked as an event and yields them.
    """

    def _pred(f):
        is_event = getattr(f, "is_event", False)
        if not is_event:
            return False

        if not f.scan:
            return False

        return True

    for _, item in inspect.getmembers(obb, predicate=_pred):
        yield (_, item)


class EventContext(object):
    """
    Represents a special context that are passed to events.
    """

    def __init__(self, cl: 'md_client.Client', shard_id: int,
                 event_name: str):
        """
        :param cl: The :class:`.Client` instance for this event context.
        :param shard_id: The shard ID this event is for.
        :param event_name: The event name for this event.
        """
        #: The :class:`.Client` instance that this event was fired under.
        self.bot = cl

        #: The shard this event was received on.
        self.shard_id = shard_id  # type: int
        #: The shard for this bot.
        self.shard_count = cl.shard_count  # type: int

        #: The event name for this event.
        self.event_name = event_name  # type: str

    @property
    def handlers(self) -> typing.List[typing.Callable[['EventContext'], None]]:
        """
        :return: A list of handlers registered for this event. 
        """
        return self.bot.events.getall(self.event_name, [])

    async def change_status(self, *args, **kwargs) -> None:
        """
        Changes the current status for this shard.
        
        This takes the same arguments as :class:`.Client.change_status`, but ignoring the shard ID.
        """
        kwargs["shard_id"] = self.shard_id
        return await self.bot.change_status(*args, **kwargs)

    @property
    def gateway(self) -> GatewayHandler:
        """
        :return: The :class:`.Gateway` that produced this event.
        """
        return self.bot.gateways[self.shard_id]



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for voice state objects.

.. currentmodule:: curious.dataclasses.voice_state
"""

from curious.dataclasses import channel as dt_channel, guild as dt_guild, member as dt_member


class VoiceState(object):
    """
    Represents the voice state of a user.
    """
    __slots__ = ("user_id", "guild_id", "channel_id", "_self_mute",
                 "_server_mute", "_self_deaf", "_server_deaf", "_bot")

    def __init__(self, **kwargs) -> None:
        self._bot = kwargs.get("client")

        #: The ID of the user for this VoiceState.
        self.user_id = int(kwargs.get("user_id", 0)) or None

        #: The ID of the guild for this VoiceState.
        self.guild_id = int(kwargs.get("guild_id", 0)) or None

        #: The ID of the channel for this VoiceState.
        self.channel_id = int(kwargs.get("channel_id", 0)) or None

        # Internal state values.
        self._self_mute = kwargs.get("self_mute", False)
        self._server_mute = kwargs.get("mute", False)
        self._self_deaf = kwargs.get("self_deaf", False)
        self._server_deaf = kwargs.get("deaf", False)

    @property
    def guild(self) -> 'dt_guild.Guild':
        """
        :return: The :class:`.Guild` associated, or None if the guild is uncached.
        """
        return self._bot.guilds.get(self.guild_id)

    @property
    def channel(self) -> 'dt_channel.Channel':
        """
        :return: The :class:`.Channel` associated, or None if the channel is uncached.
        """
        return self.guild.channels.get(self.channel_id)

    @property
    def member(self) -> 'dt_member.Member':
        """
        :return: The :class:`.Member` associated, or None.
        """
        return self.guild.members.get(self.user_id)

    @property
    def muted(self) -> bool:
        """
        :return: If this user is muted or not.
        """
        return self._server_mute or self._self_mute

    @property
    def deafened(self) -> bool:
        """
        :return: If this user is deafened or not.
        """
        return self._server_deaf or self._self_deaf

    def __repr__(self):
        return "<VoiceState user={} deaf={} mute={} channel={}>".format(self.member.user,
                                                                        self.deafened,
                                                                        self.muted,
                                                                        self.channel)

    async def mute(self):
        """
        Server mutes this member on the guild.
        """
        return await self.guild.change_voice_state(self.member, mute=True)

    async def unmute(self):
        """
        Server unmutes this member on the guild.
        """
        return await self.guild.change_voice_state(self.member, mute=False)

    async def deafen(self):
        """
        Server deafens this member on the guild.
        """
        return await self.guild.change_voice_state(self.member, deaf=True)

    async def undeafen(self):
        """
        Server undeafens this member on the guild.
        """
        return await self.guild.change_voice_state(self.member, deaf=False)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Guild objects.

.. currentmodule:: curious.dataclasses.guild
"""
import abc
import collections
import copy
import datetime
import enum
import sys
import typing
from math import ceil
from os import PathLike
from types import MappingProxyType

import multio
from dataclasses import dataclass

from curious.core.httpclient import Endpoints
from curious.dataclasses import channel as dt_channel, emoji as dt_emoji, invite as dt_invite, \
    member as dt_member, permissions as dt_permissions, role as dt_role, \
    search as dt_search, user as dt_user, voice_state as dt_vs, webhook as dt_webhook
from curious.dataclasses.bases import Dataclass
from curious.dataclasses.presence import Presence, Status
from curious.exc import CuriousError, HTTPException, HierarchyError, PermissionsError
from curious.util import AsyncIteratorWrapper, base64ify, deprecated

default_var = typing.TypeVar("T")


class MFALevel(enum.IntEnum):
    """
    Represents the MFA level of a :class:`.Guild`.
    """
    #: Used when MFA authentication is **disabled**.
    #: This means moderation actions will not require multi-factor auth.
    DISABLED = 0

    #: Used when MFA authentication is **enabled**.
    #: This means moderation actions *will* require multi-factor auth.
    ENABLED = 1


class VerificationLevel(enum.IntEnum):
    """
    Represents the verification levels for a :class:`.Guild`.
    """
    #: No verification level.
    #: All users can speak after joining immediately.
    NONE = 0

    #: Low verification level.
    #: Users must have a verified email on their account.
    LOW = 1

    #: Medium verification level.
    #: Users must have been on Discord for longer than 5 minutes.
    MEDIUM = 2

    #: High/tableflip verification level.
    #: Users must have been on the server for longer than 10 minutes.
    TABLEFLIP = 3

    #: Extreme/double tableflip verification level.
    #: Users must have a phone number associated with their account.
    DOUBLE_TABLEFLIP = 4

    def can_speak(self, member: 'dt_member.Member') -> bool:
        """
        Checks if a :class:`.Member` can speak in their :class:`.Guild`.
        
        :param member: The member to check.
        :return: True if they can speak, False if they can't.
        """
        # none always allows people to speak
        if self is VerificationLevel.NONE:
            return True

        if self is VerificationLevel.LOW:
            # can't validate, assume True
            if member.user.verified is None:
                return True

            return member.user.verified is True

        if self is VerificationLevel.MEDIUM:
            dt = datetime.datetime.now() - datetime.timedelta(minutes=5)

            # ensure their created at time is before 5 minutes before now
            if member.user.created_at < dt:
                return True

        if self is VerificationLevel.TABLEFLIP:
            dt = datetime.datetime.now() - datetime.timedelta(minutes=10)

            # ensure their joined at time is before 10 minutes before now
            if member.joined_at < dt:
                return True

        # other verification levels ???
        return True


class NotificationLevel(enum.IntEnum):
    """
    Represents the default notification level for a :class:`.Guild`.
    """
    #: All messages notify members, by default.
    ALL_MESSAGES = 0

    #: Only mentions notify members, by default.
    ONLY_MENTIONS = 1


class ContentFilterLevel(enum.IntEnum):
    """
    Represents the content filter level for a :class:`.Guild`.
    """
    #: No messages will be scanned.
    SCAN_NONE = 0

    #: Messages from users without roles will be scanned.
    SCAN_WITHOUT_ROLES = 1

    #: All messages will be scanned.
    SCAN_ALL = 2


class _WrapperBase(collections.Mapping, collections.Iterable):
    """
    Represents the base class for a wrapper object.
    """

    __slots__ = ()

    @property
    @abc.abstractmethod
    def view(self) -> 'typing.Mapping[int, Dataclass]':
        """
        Represents a read-only view for this wrapper.
        """

    def __iter__(self) -> typing.Iterator[typing.Any]:
        return iter(self.view.keys())

    def __repr__(self) -> str:
        return "<{} items='{}'>".format(type(self).__name__, self.view)


class GuildChannelWrapper(_WrapperBase):
    """
    A wrapper for channels on a guild. This provides some convenience methods which make channel
    management more fluent.
    """
    __slots__ = "_guild", "_channels"

    def __init__(self, guild: 'Guild',
                 channels: 'typing.MutableMapping[int, dt_channel.Channel]'):
        """
        :param guild: The :class:`.Guild` object that owns this wrapper.
        :param channels: The dictionary of channels that this wrapper contains.
        """
        self._guild = guild
        self._channels = channels

    @property
    def view(self) -> 'typing.Mapping[int, dt_channel.Channel]':
        """
        :return: A read-only view into the channels for this wrapper.
        """
        return MappingProxyType(self._channels)

    def __getitem__(self, key) -> 'dt_channel.Channel':
        default = object()
        got = self.get(key, default=default)
        if got is default:
            raise KeyError(key)

        return got

    def __len__(self) -> int:
        return len(self._channels)

    # overwritten methods from the abc
    def get(self, key: typing.Union[str, int], default: default_var = None) \
            -> 'typing.Union[dt_channel.Channel, default_var]':
        """
        Gets a channel by name or ID.

        :param key: The key to use. This can be the ID of the channel, or the name of the channel.
        :param default: The default value to use, if the channel cannot be found.
        :return: A :class:`.Channel`, if it was found.
        """
        if isinstance(key, int):
            return self._channels.get(key, default)
        else:
            return self._get_by_name(key, default=default)

    def _get_by_name(self, name: str, default: default_var = None) \
            -> 'typing.Union[dt_channel.Channel, default_var]':
        """
        Gets a channel by name.

        .. warning::

            This will return the first channel in the channel list. If you want to get a channel
            in a specific category, use :meth:`.Channel.get_by_name`

        :param name: The name of the channel to get.
        :param default: The default value to get, if the channel cannot be found.
        :return: A :class:`.Channel` if it can be found.
        """
        s = sorted(self._channels.values(), key=lambda c: c.position)
        try:
            return next(filter(lambda ch: ch.name == name, s))
        except StopIteration:
            return default

    async def create(self, name: str, type_: 'dt_channel.ChannelType' = None,
                     permission_overwrites: 'typing.List[dt_permissions.Overwrite]' = None,
                     *,
                     parent: 'dt_channel.Channel' = None,
                     bitrate: int = 64, user_limit: int = 0,
                     topic: str = None) -> 'dt_channel.Channel':
        """
        Creates a new channel.

        :param name: The name of the channel.
        :param type_: The :class:`.ChannelType` of the channel.
        :param permission_overwrites: The list of permission overwrites to use for this channel.

        For guild channels:

        :param parent: The parent :class:`.Channel` for this channel.

        For voice channels:

        :param bitrate: The bitrate of the channel, if it is a voice channel, in kbit/s.
        :param user_limit: The maximum number of users that can be in the channel.

        For text channels:

        :param topic: The topic of the channel, or None to set no topic.
        """
        if not self._guild.me.guild_permissions.manage_channels:
            raise PermissionsError("manage_channels")

        if type_ is None:
            type_ = dt_channel.ChannelType.TEXT

        kwargs = {
            "name": name,
            "type": type_.value,
            "permission_overwrites": permission_overwrites,
        }
        if type_ is dt_channel.ChannelType.VOICE:
            kwargs["bitrate"] = bitrate
            kwargs["user_limit"] = user_limit

        if parent is not None:
            if parent.type != dt_channel.ChannelType.CATEGORY:
                raise CuriousError("Cannot create channel with non-category parent")

            if type_.value == dt_channel.ChannelType.CATEGORY:
                raise CuriousError("Cannot create category channel with category")

            kwargs["parent_id"] = parent.id

        # create a listener so we wait for the WS before editing
        async def _listener(channel: dt_channel.Channel):
            if channel.name == name and channel.guild == self._guild:
                return True

            return False

        async with self._guild._bot.events.wait_for_manager("channel_update", _listener):
            channel_data = await self._guild._bot.http.create_channel(self._guild.id, **kwargs)
            # if it's a text channel and the topic was provided, automatically add it
            if type is dt_channel.ChannelType.TEXT and topic is not None:
                await self._guild._bot.http.edit_channel(channel_id=channel_data["id"], topic=topic)

        return self._channels[int(channel_data.get("id"))]

    def edit(self, channel: 'dt_channel.Channel', **kwargs):
        """
        Edits a channel.
        """
        if channel.id not in self._channels:
            raise CuriousError("This channel is not part of this guild")

        return channel.edit(**kwargs)

    def delete(self, channel: 'dt_channel.Channel'):
        """
        Deletes a channel.
        """
        if channel.id not in self._channels:
            raise CuriousError("This channel is not part of this guild")

        return channel.delete()


class GuildRoleWrapper(_WrapperBase):
    """
    A wrapper for roles on a guild. Contains some convenience methods that make role management
    more fluent.
    """

    __slots__ = "_guild", "_roles"

    def __init__(self, guild: 'Guild',
                 roles: 'typing.MutableMapping[int, dt_role.Role]'):
        """
        :param guild: The :class:`.Guild` object that owns this wrapper.
        :param roles: The dictionary of roles that this wrapper contains.
        """
        self._guild = guild
        self._roles = roles

    @property
    def view(self) -> 'typing.Mapping[int, dt_role.Role]':
        """
        :return: A read-only view into the channels for this wrapper.
        """
        return MappingProxyType(self._roles)

    def __getitem__(self, key) -> 'dt_role.Role':
        default = object()
        got = self.get(key, default=default)
        if got is default:
            raise KeyError(key)

        return got

    def __len__(self) -> int:
        return len(self._roles)

    # overwritten methods from the abc
    def get(self, key: typing.Union[str, int], default: default_var = None) \
            -> 'typing.Union[dt_role.Role, default_var]':
        """
        Gets a role by name or ID.

        :param key: The key to use. This can be the ID of the role, or the name of the role.
        :param default: The default value to use, if the role cannot be found.
        :return: A :class:`.Role`, if it was found.
        """
        if isinstance(key, int):
            return self._roles.get(key, default)
        else:
            return self._get_by_name(key, default=default)

    def _get_by_name(self, name: str, default: default_var = None) \
            -> 'typing.Union[dt_role.Role, default_var]':
        """
        Gets a role by name.

        :param name: The name of the channel to get.
        :param default: The default value to get, if the role cannot be found.
        :return: A :class:`.Role` if it can be found.
        """
        s = sorted(self._roles.values(), key=lambda c: c.position)
        try:
            return next(filter(lambda r: r.name == name, s))
        except StopIteration:
            return default

    async def create(self, **kwargs) -> 'dt_role.Role':
        """
        Creates a new role in this guild.

        :return: A new :class:`.Role`.
        """
        if not self._guild.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        role_obb = dt_role.Role(client=self._guild._bot,
                                **(await self._guild._bot.http.create_role(self._guild.id)))
        self._roles[role_obb.id] = role_obb
        role_obb.guild_id = self._guild.id
        return await role_obb.edit(**kwargs)

    def edit(self, role: 'dt_role.Role', **kwargs):
        """
        Edits a role.
        """
        if role.id not in self._roles:
            raise CuriousError("This role is not part of this guild")

        return role.edit(**kwargs)

    def delete(self, role: 'dt_role.Role'):
        """
        Deletes a role.
        """
        if role.id not in self._roles:
            raise CuriousError("This role is not part of this guild")

        return role.delete()


class GuildEmojiWrapper(_WrapperBase):
    """
    Wrapper for emoji objects for a guild.
    """
    __slots__ = "_guild", "_emojis"

    def __init__(self, guild: 'Guild',
                 emojis: 'typing.MutableMapping[int, dt_emoji.Emoji]'):
        """
        :param guild: The :class:`.Guild` object that owns this wrapper.
        :param emojis: The dictionary of emojis that this wrapper contains.
        """
        self._guild = guild
        self._emojis = emojis

    @property
    def view(self) -> 'typing.Mapping[int, dt_emoji.Emoji]':
        """
        :return: A read-only view into the channels for this wrapper.
        """
        return MappingProxyType(self._emojis)

    def __getitem__(self, key) -> 'dt_emoji.Emoji':
        return self._emojis[key]

    def __len__(self) -> int:
        return len(self._emojis)

    async def create(self, *,
                     name: str, image_data: typing.Union[str, bytes],
                     roles: 'typing.List[dt_role.Role]' = None) -> 'dt_emoji.Emoji':
        """
        Creates a new emoji in this guild.

        :param name: The name of the emoji.
        :param image_data: The bytes image data or the str base64 data for the emoji.
        :param roles: A list of roles this emoji is locked to.
        :return: The :class:`.Emoji` created.
        """
        if isinstance(image_data, bytes):
            image_data = base64ify(image_data)

        if roles is not None:
            roles = [r.id for r in roles]

        emoji_data = await self._guild._bot.http.create_guild_emoji(self._guild.id,
                                                                    name=name,
                                                                    image_data=image_data,
                                                                    roles=roles)
        emoji = dt_emoji.Emoji(**emoji_data, client=self._guild._bot)
        return emoji


class GuildBan:
    """
    Represents a ban in a guild.
    """
    #: The reason for the ban.
    reason: str

    #: The victim of the ban.
    victim: 'dt_user.User'


if 'sphinx' in sys.modules:
    # fuck you
    pass
else:
    GuildBan = dataclass(GuildBan, frozen=True)


class GuildBanContainer(object):
    """
    A container for guild bans.
    """

    def __init__(self, guild: 'Guild'):
        self._guild = guild

    async def __aiter__(self) -> 'typing.AsyncGenerator[GuildBan]':
        if not self._guild.me.guild_permissions.ban_members:
            raise PermissionsError("ban_members")

        bans = await self._guild._bot.http.get_bans(self._guild.id)

        for ban in bans:
            user_data = ban.get("user", None)
            if user_data is None:
                continue

            user = self._guild._bot.state.make_user(user_data)
            self._guild._bot.state._check_decache_user(user.id)
            ban = GuildBan(reason=ban.get("reason", None), user=user)
            yield ban

    async def add(self, victim: 'typing.Union[dt_user.User, dt_member.Member]', *,
                  delete_message_days: int, reason: str = None) -> GuildBan:
        """
        Bans somebody from the guild.

        This can either ban a :class:`.Member`, in which they must be in the guild.
        Or this can ban a :class:`.User`, which does not need to be in the guild.

        Example for banning a member:

        .. code:: python

            member = guild.members[66237334693085184]
            await guild.ban(member)

        Example for banning a user:

        .. code:: python

            user = await client.get_user(66237334693085184)
            await guild.ban(user)

        :param victim: The :class:`.Member` or :class:`.User` object to ban.
        :param delete_message_days: The number of days to delete messages.
        :param reason: The reason given for banning.
        """
        if not self._guild.me.guild_permissions.ban_members:
            raise PermissionsError("ban_members")

        if isinstance(victim, dt_member.Member):
            if self._guild.owner == victim:
                raise HierarchyError("Cannot ban the owner")

            if victim.guild_id != self._guild.id:
                raise ValueError("Member must be from this guild (try `member.user` instead!)")

            if victim.top_role >= self._guild.me.top_role:
                raise HierarchyError("Top role is equal to or lower than victim's top role")

            victim_user = victim.user
            victim_id = victim.user.id

        elif isinstance(victim, dt_user.User):
            victim_user = victim
            victim_id = victim.id

        else:
            raise TypeError("Victim must be a Member or a User")

        await self._guild._bot.http.ban_user(guild_id=self._guild.id, user_id=victim_id,
                                             delete_message_days=delete_message_days,
                                             reason=reason)
        return GuildBan(reason=reason, victim=victim_user)

    async def ban(self, *args, **kwargs) -> 'GuildBan':
        """
        Shortcut for :meth:`.GuildBanWrapper.add`.
        """
        return await self.add(*args, **kwargs)

    async def remove(self, user: 'dt_user.User', *,
                     reason: str = None) -> None:
        """
        Unbans a user from this guild.
        Example for unbanning the first banned user:

        .. code-block:: python3

            user = next(await guild.get_bans())
            await guild.unban(user)

        To unban an arbitrary user, use :meth:`.Client.get_user`.

        .. code-block:: python3

            user = await client.get_user(66237334693085184)
            await guild.unban(user)

        .. note::

            This does not take :class:`.Member` objects, as members cannot be in a guild and
            banned from the guild.

        :param user: The :class:`.User` to forgive and unban.
        :param reason: The reason given for unbanning.
        """
        if not self._guild.me.guild_permissions.ban_members:
            raise PermissionsError("ban_members")

        forgiven_id = user.id

        await self._guild._bot.http.unban_user(self._guild.id, forgiven_id, reason=reason)

    async def unban(self, *args, **kwargs) -> None:
        """
        Shortcut for :meth:`.GuildBanWrapper.remove`.
        """
        return await self.remove(*args, **kwargs)

    async def flatten(self) -> 'typing.List[GuildBan]':
        """
        Gets all the bans for this guild.
        """
        return [ban async for ban in self]


class Guild(Dataclass):
    """
    Represents a guild object on Discord.
    """

    __slots__ = (
        "id", "unavailable", "name", "afk_timeout", "region",
        "mfa_level", "verification_level", "notification_level", "content_filter_level", "features",
        "shard_id", "_roles", "_members", "_channels", "_emojis", "member_count", "_voice_states",
        "_large", "_chunks_left", "_finished_chunking", "icon_hash", "splash_hash",
        "owner_id", "afk_channel_id", "system_channel_id", "widget_channel_id",
        "voice_client",
        "channels", "roles", "emojis", "bans",
    )

    valid_embed_styles = {'banner1', 'banner3', 'banner2', 'shield', 'banner4'}

    def __init__(self, bot, **kwargs) -> None:
        super().__init__(kwargs.get("id"), bot)

        #: If the guild is unavailable or not.
        #: If this is True, many fields return `None`.
        self.unavailable = kwargs.get("unavailable", False)

        # Placeholder values.
        #: The name of this guild.
        self.name = None  # type: str

        #: The icon hash of this guild.
        #: Used to construct the icon URL later.
        self.icon_hash = None  # type: str

        #: The splash hash of this guild.
        #: Used to construct the splash URL later.
        self.splash_hash = None  # type: str

        #: The AFK channel ID of this guild.
        self.afk_channel_id = None  # type: int

        #: The ID of the system channel for this guild.
        #: This is where welcome messages and the likes are sent.
        #: Effective replacement for default channel for bots.
        self.system_channel_id = None  # type: int

        #: The widget channel ID for this guild.
        self.widget_channel_id = None  # type: int

        #: The owner ID of this guild.
        self.owner_id = None  # type: int

        #: The AFK timeout for this guild. None if there's no AFK timeout.
        self.afk_timeout = None  # type: int

        #: The voice region of this guild.
        self.region = None  # type: str

        #: The features this guild has.
        self.features = None  # type: typing.List[str]

        #: The MFA level of this guild.
        self.mfa_level = MFALevel.DISABLED
        #: The verification level of this guild.
        self.verification_level = VerificationLevel.NONE
        #: The notification level of this guild.
        self.notification_level = NotificationLevel.ALL_MESSAGES
        #: The content filter level of this guild.
        self.content_filter_level = ContentFilterLevel.SCAN_NONE

        #: The shard ID this guild is associated with.
        self.shard_id = None  # type: int

        #: The roles that this guild has.
        self._roles = {}
        #: The members of this guild.
        self._members = {}
        #: The channels of this guild.
        self._channels = {}
        #: The emojis that this guild has.
        self._emojis = {}
        #: The voice states that this guild has.
        self._voice_states = {}

        #: The number of numbers this guild has.
        #: This is automatically updated.
        self.member_count = 0  # type: int

        #: Is this guild a large guild according to Discord?
        self._large = None  # type: bool

        #: Has this guild finished chunking?
        self._finished_chunking = multio.Event()
        self._chunks_left = 0

        #: The current voice client associated with this guild.
        self.voice_client = None

        #: The :class:`.GuildChannelWrapper` that wraps the channels in this Guild.
        self.channels = GuildChannelWrapper(self, self._channels)
        #: The :class:`.GuildRoleWrapper` that wraps the roles in this Guild.
        self.roles = GuildRoleWrapper(self, self._roles)
        #: The :class:`.GuildEmojiWrapper` that wraps the emojis in this Guild.
        self.emojis = GuildEmojiWrapper(self, self._emojis)
        #: The :class:`.GuildBanContainer` for this Guild.
        self.bans = GuildBanContainer(self)

        if kwargs:
            self.from_guild_create(**kwargs)

    def _copy(self) -> 'Guild':
        return copy.copy(self)

    def __repr__(self) -> str:
        return "<Guild id='{}' name='{}' members='{}'>".format(self.id, self.name,
                                                               self.member_count)

    def __str__(self) -> str:
        return repr(self)

    @property
    def members(self) -> 'typing.Mapping[int, dt_member.Member]':
        """
        :return: A mapping of :class:`.Member` that represent members on this guild.
        """
        return MappingProxyType(self._members)

    @property
    def owner(self) -> 'typing.Union[dt_member.Member, None]':
        """
        :return: A :class:`.Member` object that represents the owner of this guild.
        """
        try:
            return self._members[self.owner_id]
        except KeyError:
            return None

    @property
    def me(self) -> 'typing.Union[dt_member.Member, None]':
        """
        :return: A :class:`.Member` object that represents the current user in this guild.
        """
        try:
            return self._members[self._bot.user.id]
        except KeyError:
            return None

    @property
    def default_role(self) -> 'typing.Union[dt_role.Role, None]':
        """
        :return: A :class:`.Role` that represents the default role of this guild.
        """
        try:
            return self.roles[self.id]
        except KeyError:
            return None

    @property
    def system_channel(self) -> 'typing.Union[dt_channel.Channel, None]':
        """
        :return: A :class:`.Channel` that represents the system channel for this guild.
        """
        try:
            return self._channels[self.system_channel_id]
        except KeyError:
            return None

    @property
    def afk_channel(self) -> 'typing.Union[dt_channel.Channel, None]':
        """
        :return: A :class:`.Channel` representing the AFK channel for this guild.
        """
        try:
            return self._channels[self.afk_channel_id]
        except IndexError:
            return None

    @property
    def embed_url(self) -> str:
        """
        Gets the default embed url for this guild.
        If the widget is not enabled, this endpoint will 404.
        
        :return: The embed URL for this guild. 
        """
        return (Endpoints.GUILD_BASE + "/embed.png").format(guild_id=self.id)

    # for parity with inviteguild
    @property
    def presence_count(self) -> int:
        """
        :return: The number of members with a non-Invisible presence. 
        """
        return sum(1 for member in self._members.values() if member.status is not Status.OFFLINE)

    # Presence methods
    def members_with_status(self, status: Status) \
            -> 'typing.Generator[dt_member.Member, None, None]':
        """
        A generator that returns the members that match the specified status.
        """
        for member in self.members.values():
            if member.status == status:
                yield member

    @property
    def online_members(self) -> 'typing.Generator[dt_member.Member, None, None]':
        """
        :return: A generator of online :class:`.Member` objects.
        """
        return self.members_with_status(Status.ONLINE)

    @property
    def idle_members(self) -> 'typing.Generator[dt_member.Member, None, None]':
        """
        :return: A generator of idle :class:`.Member` objects.
        """
        return self.members_with_status(Status.IDLE)

    @property
    def dnd_members(self) -> 'typing.Generator[dt_member.Member, None, None]':
        """
        :return: A generator of DnD :class:`.Member` objects.
        """
        return self.members_with_status(Status.DND)

    @property
    def offline_members(self) -> 'typing.Generator[dt_member.Member, None, None]':
        """
        :return: A generator of offline/invisible :class:`.Member` objects.
        """
        return self.members_with_status(Status.OFFLINE)

    @property
    def search(self) -> 'dt_search.SearchQuery':
        """
        :return: A :class:`.SearchQuery` that can be used to search this guild's messages.
        """
        return dt_search.SearchQuery(guild=self)

    def get_embed_url(self, *, style: str = "banner1") -> str:
        """
        Gets an embed URL for this guild in a specified style.
        
        :param style: The style to get. 
        :return: The embed URL for this guild.
        """
        if style not in self.valid_embed_styles:
            raise ValueError("Style must be in {}".format(self.valid_embed_styles))

        return self.embed_url + "?style={}".format(style)

    def search_for_member(self, *, name: str = None, discriminator: str = None,
                          full_name: str = None):
        """
        Searches for a member.

        :param name: The username or nickname of the member.
        :param discriminator: The discriminator of the member.
        :param full_name: The full name (i.e. username#discrim) of the member. Optional; will be \
            split up into the correct parameters.

        .. warning::

            Using a username and discriminator pair is most accurate when finding a user; a
            nickname pair or not providing one of the arguments might not find the right member.

        :return: A :class:`.Member` that matched, or None if no matches were found.
        """
        if full_name is not None:
            if "#" in full_name:
                sp = full_name.split("#", 1)
                return self.search_for_member(name=sp[0], discriminator=sp[1])
            else:
                # usually a mistake
                return self.search_for_member(name=full_name)

        # coerce into a proper string
        if isinstance(discriminator, int):
            discriminator = "{:04d}".format(discriminator)

        for member in self._members.values():
            # ensure discrim matches first
            if discriminator is not None and discriminator != member.user.discriminator:
                continue

            if member.user.username == name:
                return member

            if member.nickname == name:
                return member

    @deprecated(since="0.7.0", see_instead=search_for_member, removal="0.9.0")
    def find_member(self, search_str: str) -> 'dt_member.Member':
        """
        Attempts to find a member in this guild by name#discrim.
        This will also search nicknames.

        The discriminator is optional, but if provided allows better matching.

        :param search_str: The name#discrim pair to search for.
        :return: A :class:`.Member` object that represents the member, or None if no member \
                could be found.
        """
        sp = search_str.rsplit("#", 1)
        if len(sp) == 1:
            # Member name only :(
            predicate = lambda member: member.user.name == sp[0] or member.nickname == sp[0]
        else:
            # Discriminator too!
            # Don't check nicknames for this.
            predicate = lambda member: member.user.name == sp[0] \
                                       and member.user.discriminator == sp[1]

        filtered = filter(predicate, self.members.values())
        return next(filtered, None)

    # creation methods
    def start_chunking(self) -> None:
        """
        Marks a guild to start guild chunking.
        
        This will clear the chunking event, and calculate the number of member chunks required.
        """
        self._finished_chunking.clear()
        self._chunks_left = ceil(self.member_count / 1000)

    async def wait_until_chunked(self) -> None:
        """
        Waits until the guild has finished chunking.

        Useful for when you join a big guild.
        """
        await self._finished_chunking.wait()

    def _handle_member_chunk(self, members: list):
        """
        Handles a chunk of members.
        
        :param members: A list of member data dictionaries as returned from Discord.
        """
        if self._chunks_left >= 1:
            # We have a new chunk, so decrement the number left.
            self._chunks_left -= 1

        for member_data in members:
            member_id = int(member_data["user"]["id"])
            if member_id in self._members:
                member_obj = self._members[member_id]
            else:
                member_obj = dt_member.Member(self._bot, **member_data)
                self._members[member_obj.id] = member_obj

            member_obj.nickname = member_data.get("nick", member_obj.nickname)
            member_obj.guild_id = self.id

    def _handle_emojis(self, emojis: typing.List[dict]):
        """
        Handles the emojis for this guild.
        
        :param emojis: A list of emoji objects from Discord.
        """
        for emoji in emojis:
            emoji_obj = dt_emoji.Emoji(**emoji, client=self._bot)
            self._emojis[emoji_obj.id] = emoji_obj
            emoji_obj.guild_id = self.id

    def from_guild_create(self, **data: dict) -> 'Guild':
        """
        Populates the fields from a GUILD_CREATE event.

        :param data: The GUILD_CREATE data to use.
        """
        self.unavailable = data.get("unavailable", False)

        if self.unavailable:
            # We can't use any of the extra data here, so don't bother.
            return self

        self.name = data.get("name")  # type: str
        self.icon_hash = data.get("icon")  # type: str
        self.splash_hash = data.get("splash")  # type: str
        self.owner_id = int(data.get("owner_id", 0)) or None  # type: int
        self._large = data.get("large", None)
        self.features = data.get("features", [])
        self.region = data.get("region")

        afk_channel_id = data.get("afk_channel_id", 0)
        if afk_channel_id:
            afk_channel_id = int(afk_channel_id)

        self.afk_channel_id = afk_channel_id
        self.afk_timeout = data.get("afk_timeout")

        self.verification_level = VerificationLevel(data.get("verification_level", 0))
        self.mfa_level = MFALevel(data.get("mfa_level", 0))
        self.notification_level = NotificationLevel(data.get("default_message_notifications", 0))
        self.content_filter_level = ContentFilterLevel(data.get("explicit_content_filter", 0))

        self.member_count = data.get("member_count", 0)

        # Create all the Role objects for the server.
        for role_data in data.get("roles", []):
            role_obj = dt_role.Role(self._bot, **role_data)
            role_obj.guild_id = self.id
            self._roles[role_obj.id] = role_obj

        # Create all the Member objects for the server.
        self._handle_member_chunk(data.get("members", []))

        for presence in data.get("presences", []):
            member_id = int(presence["user"]["id"])
            member_obj = self._members.get(member_id)

            if not member_obj:
                continue

            member_obj.presence = Presence(**presence)

        # Create all of the channel objects.
        for channel_data in data.get("channels", []):
            channel_obj = dt_channel.Channel(self._bot, **channel_data)
            self._channels[channel_obj.id] = channel_obj
            channel_obj.guild_id = self.id
            channel_obj._update_overwrites(channel_data.get("permission_overwrites", []), )

        # Create all of the voice states.
        for vs_data in data.get("voice_states", []):
            user_id = int(vs_data.get("user_id", 0))
            member = self.members.get(user_id)
            if not member:
                # o well
                continue

            voice_state = dt_vs.VoiceState(**vs_data, client=self._bot)
            self._voice_states[voice_state.user_id] = voice_state

            vs_channel = self._channels.get(int(vs_data.get("channel_id", 0)))
            if vs_channel is not None:
                voice_state.channel_id = vs_channel.id
                voice_state.guild_id = self.id

        # delegate to other function
        self._handle_emojis(data.get("emojis", []))

    @property
    def large(self) -> bool:
        """
        :return: If this guild is large or not (>= 250 members).
        """
        if self._large is not None:
            return self._large

        return self.member_count >= 250

    @property
    def invites(self) -> 'typing.AsyncIterator[dt_invite.Invite]':
        """
        :return: A class:`.AsyncIteratorWrapper` that yields :class:`.Invite` objects for this
            guild.
        """
        return AsyncIteratorWrapper(self.get_invites)

    @property
    def icon_url(self) -> str:
        """
        :return: The icon URL for this guild, or None if one isn't set.
        """
        if self.icon_hash:
            return "https://cdn.discordapp.com/icons/{}/{}.webp".format(self.id, self.icon_hash)

    @property
    def splash_url(self) -> str:
        """
        :return: The splash URL for this guild, or None if one isn't set.
        """
        if self.splash_hash:
            return "https://cdn.discordapp.com/splashes/{}/{}.webp".format(self.id,
                                                                           self.splash_hash)

    # Guild methods.
    async def leave(self) -> None:
        """
        Leaves the guild.
        """
        await self._bot.http.leave_guild(self.id)

    # async def connect_to_voice(self, channel: 'dt_channel.Channel') -> 'voice_client.VoiceClient':
    #     """
    #     Connects to a voice channel in this guild.
    #
    #     :param channel: The :class:`.Channel` to connect to.
    #     :return: The :class:`VoiceClient` that was connected to this guild.
    #     """
    #     if voice_client is None:
    #         raise RuntimeError("Cannot connect to voice - voice support is not installed")
    #
    #     if channel.guild != self:
    #         raise CuriousError("Cannot use channel from a different guild")
    #
    #     if self.voice_client is not None and self.voice_client.open:
    #         raise CuriousError("Voice client already exists in this guild")
    #
    #     gw = self._bot._gateways[self.shard_id]
    #     self.voice_client = await voice_client.VoiceClient.create(self._bot, gw, channel)
    #     await self.voice_client.connect()
    #     return self.voice_client

    async def get_invites(self) -> 'typing.List[dt_invite.Invite]':
        """
        Gets the invites for this guild.
        :return: A list :class:`.Invite` objects.
        """
        invites = await self._bot.http.get_invites_for(self.id)
        invites = [dt_invite.Invite(self._bot, **i) for i in invites]

        try:
            invite = await self.get_vanity_invite()
            invites.insert(0, invite)
        except (CuriousError, HTTPException):
            pass

        return invites

    async def kick(self, victim: 'dt_member.Member'):
        """
        Kicks somebody from the guild.

        :param victim: The :class:`.Member` to kick.
        """
        if not self.me.guild_permissions.kick_members:
            raise PermissionsError("kick_members")

        if victim.guild != self:
            raise ValueError("Member must be from this guild (try `member.user` instead)")

        if victim.top_role >= self.me.top_role:
            raise HierarchyError("Top role is equal to or lower than victim's top role")

        victim_id = victim.user.id

        await self._bot.http.kick_member(self.id, victim_id)

    @deprecated(since="0.7.0", see_instead=GuildBanContainer.add, removal="0.9.0")
    async def ban(self, victim: 'typing.Union[dt_member.Member, dt_user.User]', *,
                  delete_message_days: int = 7) -> GuildBan:
        """
        Bans somebody from the guild.

        This can either ban a :class:`.Member`, in which they must be in the guild.
        Or this can ban a :class:`.User`, which does not need to be in the guild.

        Example for banning a member:

        .. code:: python

            member = guild.members[66237334693085184]
            await guild.ban(member)

        Example for banning a user:

        .. code:: python

            user = await client.get_user(66237334693085184)
            await guild.ban(user)

        :param victim: The :class:`.Member` or :class:`.User` object to ban.
        :param delete_message_days: The number of days to delete messages.
        """
        return await self.bans.add(victim, delete_message_days=delete_message_days)

    @deprecated(since="0.7.0", see_instead=GuildBanContainer.remove, removal="0.9.0")
    async def unban(self, user: 'dt_user.User') -> None:
        """
        Unbans a user from this guild.
        Example for unbanning the first banned user:

        .. code:: python

            user = next(await guild.get_bans())
            await guild.unban(user)

        To unban an arbitrary user, use :meth:`.Client.get_user`.

        .. code:: python

            user = await client.get_user(66237334693085184)
            await guild.unban(user)

        .. note::

            This does not take :class:`.Member` objects, as members cannot be in a guild and
            banned from the guild.

        :param user: The :class:`.User` to forgive and unban.
        """
        return await self.bans.remove(user)

    async def get_webhooks(self) -> 'typing.List[dt_webhook.Webhook]':
        """
        Gets the webhooks for this guild.

        :return: A list of :class:`.Webhook` objects for the guild.
        """
        webhooks = await self._bot.http.get_webhooks_for_guild(self.id)
        obbs = []

        for webhook in webhooks:
            obbs.append(self._bot.state.make_webhook(webhook))

        return obbs

    async def delete_webhook(self, webhook: 'dt_webhook.Webhook'):
        """
        Deletes a webhook in this guild.

        :param webhook: The :class:`.Webhook` to delete.
        """
        if not self.me.guild_permissions.manage_webhooks:
            raise PermissionsError("manage_webhooks")

        await self._bot.http.delete_webhook(webhook.id)

    async def change_role_positions(self,
                                    roles: 'typing.Union[typing.Dict[dt_role.Role, int], '
                                           'typing.List[typing.Tuple[dt_role.Role, int]]]'):
        """
        Changes the positions of a mapping of roles.

        :param roles: A dict or iterable of two-item tuples of new roles that is in the format of \
            (role, position).
        """
        if not self.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        if isinstance(roles, dict):
            roles = roles.items()

        to_send = []
        for r, new_position in roles:
            if new_position >= self.me.top_role.position:
                raise HierarchyError("Cannot move role above our top role")

            to_send.append((str(r.id), new_position))

        to_send = [(str(r.id), new_position) for (r, new_position) in roles]
        await self._bot.http.edit_role_positions(to_send)

    async def change_voice_state(self, member: 'dt_member.Member', *,
                                 deaf: bool = None, mute: bool = None):
        """
        Changes the voice state of a member.

        :param member: The :class:`.Member` to change the voice state of.
        :param deaf: Should this member be deafened?
        :param mute: Should this member be muted?
        """
        if member.voice is None:
            raise CuriousError("Cannot change voice state of member not in voice")

        await self._bot.http.edit_member_voice_state(self.id, member.id, deaf=deaf, mute=mute)
        return member.voice

    async def modify_guild(self, *, afk_channel: 'dt_channel.Channel' = None,
                           verification_level: VerificationLevel = None,
                           content_filter_level: ContentFilterLevel = None,
                           notification_level: NotificationLevel = None,
                           **kwargs):
        """
        Edits this guild.

        For a list of available arguments, see 
        https://discordapp.com/developers/docs/resources/guild#modify-guild.
        
        :param afk_channel: The :class:`.Channel` that represents the AFK voice channel.
        :param verification_level: The :class:`.VerificationLevel` to use for this guild.
        :param content_filter_level: The :class:`.ContentFilterLevel` to use for this guild.
        :param notification_level: The :class:`.NotificationLevel` to use for this guild.
        """
        if not self.me.guild_permissions.manage_server:
            raise PermissionsError("manage_server")

        if afk_channel is not None:
            kwargs["afk_channel_id"] = afk_channel.id

        if verification_level is not None:
            kwargs["verification_level"] = verification_level.value

        if notification_level is not None:
            kwargs["default_message_notifications"] = notification_level.value

        if content_filter_level is not None:
            kwargs["explicit_content_filter"] = content_filter_level.value

        await self._bot.http.edit_guild(self.id, **kwargs)
        return self

    async def change_icon(self, icon_content: bytes):
        """
        Changes the icon for this guild.

        :param icon_content: The bytes that represent the icon of the guild.
        """
        if not self.me.guild_permissions.manage_server:
            raise PermissionsError("manage_server")

        image = base64ify(icon_content)
        await self._bot.http.edit_guild(self.id,
                                        icon_content=image)

    async def upload_icon(self, path: PathLike):
        """
        Uploads a new icon for the guild.

        :param path: A path-like object to use to upload.
        """
        with open(path, 'rb') as f:
            return await self.change_icon(f.read())

    async def get_widget_info(self) -> 'typing.Tuple[bool, ' \
                                       'typing.Union[None, dt_channel.Channel]]':
        """
        Gets the widget info for the current guild.
        
        :return: A two-item tuple: If this widget is enabled, and the channel the widget has an \ 
            invite for. 
        """
        info = await self._bot.http.get_widget_status(self.id)
        return info.get("enabled", False), self.channels.get(int(info.get("channel_id", 0)))

    async def edit_widget(self, *,
                          status: bool = None, channel: 'dt_channel.Channel' = -1):
        """
        Edits the widget for this guild.
        
        :param status: The status of this widget: True or False.
        :param channel: The channel object to set the instant invite to.
        """
        if channel is None:
            channel_id = None
        elif channel == -1:
            channel_id = 0
        else:
            channel_id = channel.id

        await self._bot.http.edit_widget(self.id, enabled=status, channel_id=channel_id)

    async def get_vanity_invite(self) -> 'typing.Union[None, dt_invite.Invite]':
        """
        Gets the vanity :class:`.Invite` for this guild.

        :return: The :class:`.Invite` that corresponds with this guild, if it has one.
        """
        if 'vanity-url' not in self.features:
            return None

        try:
            resp = await self._bot.http.get_vanity_url(self.id)
        except HTTPException as e:
            if e.error_code != 50020:
                raise
            else:
                return None

        code = resp.get("code", None)
        if code is None:
            return None

        invite_data = await self._bot.http.get_invite(code)
        invite = dt_invite.Invite(self._bot, **invite_data)

        return invite

    async def set_vanity_invite(self, url: str) -> 'typing.Union[dt_invite.Invite, None]':
        """
        Sets the vanity :class:`.Invite` for this guild.

        :param url: The code to use for this guild.
        :return: The :class:`.Invite` produced.
        """
        if 'vanity-url' not in self.features:
            raise CuriousError("This guild has no vanity URL")

        try:
            resp = await self._bot.http.edit_vanity_url(self.id, url)
        except HTTPException as e:
            if e.error_code != 50020:
                raise

            raise CuriousError("This guild has no vanity URL")

        code = resp.get("code", None)
        if code is None:
            return None

        invite_data = await self._bot.http.get_invite(code)
        invite = dt_invite.Invite(self._bot, **invite_data)

        return invite



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Misc utilities shared throughout the library.

.. currentmodule:: curious.util
"""
import base64
import collections
import datetime
import functools
import imghdr
import inspect
import textwrap
import types
import warnings
from typing import Any, Awaitable, Callable, Coroutine, List, Union

import multio
from multidict import MultiDict

NO_ITEM = object()


def remove_from_multidict(d: MultiDict, key: str, item: Any):
    """
    Removes an item from a multidict key.
    """
    # works by popping all, removing, then re-adding into
    i = d.popall(key, [])
    if item in i:
        i.remove(item)

    for n in i:
        d.add(key, n)

    return d


class AsyncIteratorWrapper(collections.AsyncIterator):
    """
    Wraps a coroutine that returns a sequence of items into something that can iterated over
    asynchronously.
    
    .. code-block:: python3
    
        async def a():
            # ... some long op
            return [..., ..., ...]
         
        it = AsyncIteratorWrapper(a)
        
        async for item in it:
            print(item)
    
    """

    def __init__(self, coro: Callable[[], Union[Awaitable[List[Any]], Coroutine[None, None, Any]]]):
        self.coro = coro

        self.items = collections.deque()

        self._filled = False

    async def _fill(self) -> None:
        self.items.extend(await self.coro())
        self._filled = True

    async def __anext__(self) -> Any:
        if not self._filled:
            await self._fill()

        try:
            return self.items.popleft()
        except IndexError:
            raise StopAsyncIteration

    # helper methods
    async def next(self, default=NO_ITEM) -> Any:
        """
        Gets the next item from this iterable.
        """
        try:
            return await self.__anext__()
        except StopAsyncIteration:
            if default == NO_ITEM:
                raise

            return default

    async def all(self) -> List[Any]:
        """
        Gets a flattened list of items from this iterator.
        """
        items = []

        async for item in self:
            items.append(item)

        return items


def base64ify(image_data: bytes):
    """
    Base64-ifys an image to send to discord.

    :param image_data: The data of the image to use.
    :return: A string containing the encoded image.
    """
    # Convert the avatar to base64.
    mimetype = imghdr.what(None, image_data)
    if not mimetype:
        raise ValueError("Invalid image type")

    b64_data = base64.b64encode(image_data).decode()
    return "data:{};base64,{}".format(mimetype, b64_data)


def to_datetime(timestamp: str) -> datetime.datetime:
    """
    Converts a Discord-formatted timestamp to a datetime object.

    :param timestamp: The timestamp to convert.
    :return: The :class:`datetime.datetime` object that corresponds to this datetime.
    """
    if timestamp is None:
        return

    if timestamp.endswith("+00:00"):
        timestamp = timestamp[:-6]

    try:
        return datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f")
    except ValueError:
        # wonky datetimes
        return datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S")


def replace_quotes(item: str) -> str:
    """
    Replaces the quotes in a string, but only if they are un-escaped.
    
    .. code-block:: python3
    
        some_weird_string = r'"this is quoted and removed" but \" that was kept and this isn't \\"'
        replace_quotes(some_weird_string)  # 'this is quoted and removed but " that was kept but
        this isnt \\'

    :param item: The string to scan.
    :return: The string, with quotes replaced.
    """
    # A list is used because it can be appended easily.
    final_str_arr = []

    for n, char in enumerate(item):
        # only operate if the previous char actually exists
        if n - 1 < 0:
            if char != '"':
                final_str_arr.append(char)

            continue

        # Complex quoting rules!
        # If it's a SINGLE backslash, don't append it.
        # If it's a double backslash, append it.
        if char == '\\':
            if item[n - 1] == "\\":
                # double backslash, append it
                final_str_arr.append(char)

            continue

        if char == '"':
            # check to see if it's escaped
            if item[n - 1] == '\\':
                # if the last char on final_str_arr is NOT a backslash, we want to keep it.
                if len(final_str_arr) > 0 and final_str_arr[-1] != '\\':
                    final_str_arr.append('"')

            continue

        # None of the above were hit, so add it anyway and continue.
        final_str_arr.append(char)

    return "".join(final_str_arr)


def _traverse_stack_for(t: type):
    """
    Traverses the stack for an object of type ``t``.

    :param t: The type of the object.
    :return: The object, if found.
    """
    for fr in inspect.stack():
        frame = fr.frame
        try:
            locals = frame.locals
        except AttributeError:
            # idk
            continue
        else:
            for object in locals.values():
                if type(object) is t:
                    return object
        finally:
            # prevent reference cycles
            del fr


async def coerce_agen(gen):
    """
    Coerces an async generator into a list.
    """
    results = []
    async with multio.asynclib.finalize_agen(gen) as agen:
        async for i in agen:
            results.append(i)

    return results


def subclass_builtin(original: type):
    """
    Subclasses an immutable builtin, providing method wrappers that return the subclass instead
    of the original.
    """

    def get_wrapper(subclass, func):

        @functools.wraps(func)
        def __inner_wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            new = subclass(result)

            # copy the parent dataclass if we need to
            if 'parent' in self.__dict__:
                new.__dict__['parent'] = self.__dict__['parent']

            return new

        return __inner_wrapper

    def wrapper(subclass):
        if hasattr(subclass, "__slots__"):
            raise RuntimeError("Cannot fix a slotted class")

        # get all of the methods on the original andparse the docstring; these are in a
        # well-defined format for builtins
        for meth_name in dir(original):
            func = getattr(subclass, meth_name)

            # only overwrite doc'd functions
            if not func.__doc__:
                continue

            sig = func.__doc__.split("\n")[0]

            try:
                rtype: str = sig.split("->")[1]
            except IndexError:
                continue

            # check if it matches our real method
            rtype = rtype.lstrip().rstrip()
            if rtype == original.__name__:
                # make a new function wrapper, which returns the appropriate type
                # then add it to our subclass
                wrapper = get_wrapper(subclass, func)
                setattr(subclass, meth_name, wrapper)

        return subclass

    return wrapper


class CuriousDeprecatedWarning(FutureWarning):
    """
    Warned when a function is deprecated.
    """


def deprecated(*, since: str, see_instead, removal: str):
    """
    Marks a method as deprecated.

    :param since: The version since this is deprecated.
    :param see_instead: What method to see instead.
    :param removal: The version this function will be removed at.
    """

    # TODO: In 3.7, the globals mess probably won't be needed.
    def inner(func):
        # calculate a new doc
        nonlocal see_instead
        if not isinstance(see_instead, str):
            qualname = see_instead.__qualname__
            mod = inspect.getmodule(see_instead).__name__

            # eat curious defines
            if mod.startswith("curious."):
                mod = ""

            # check for classes
            if '.' in qualname:
                see_instead = f":meth:`{mod}.{qualname}`"
            else:
                see_instead = f":func:`{mod}.{qualname}`"

        doc = inspect.getdoc(func)
        if doc is not None:
            original_doc = textwrap.dedent(func.__doc__)
            func.__doc__ = f"**This function is deprecated since {since}.** " \
                           f"See :meth:`.{see_instead}` instead.  \n" \
                           f"It will be removed at version {removal}.\n\n" \
                           f"{original_doc}"

        def wrapper(*args, **kwargs):
            warnings.warn(f"    This function is deprecated since {since}. "
                          f"    See '{see_instead}' instead.",
                          category=CuriousDeprecatedWarning, stacklevel=2)
            return func(*args, **kwargs)

        # HACKY METAPROGRAMMING
        new_globals = {**func.__globals__}
        new_globals.update(wrapper.__globals__)

        new_wrapper = types.FunctionType(
            wrapper.__code__, new_globals,
            name=wrapper.__name__, argdefs=wrapper.__defaults__,
            closure=wrapper.__closure__
        )
        new_wrapper = functools.update_wrapper(new_wrapper, func)

        new_wrapper.deprecated = True
        new_wrapper.__doc__ = inspect.getdoc(func)
        return new_wrapper

    return inner


def safe_generator(cbl):
    # only wrap if we have curio
    try:
        from curio.meta import safe_generator
        return safe_generator(cbl)
    except ModuleNotFoundError:
        return cbl


def _ad_getattr(self, key: str):
    try:
        return self[key]
    except KeyError as e:
        raise AttributeError(key) from e


attrdict = type("attrdict", (dict,), {
    "__getattr__": _ad_getattr,
    "__setattr__": dict.__setitem__,
    "__doc__": "A dict that allows attribute access as well as item access for "
               "keys."
})



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Status objects.

.. currentmodule:: curious.dataclasses.presence
"""

import enum
from typing import List


class Status(enum.Enum):
    """
    Represents a Member's status.
    """
    #: Corresponds to online (green dot).
    ONLINE = 'online'

    #: Corresponds to offline (gray dot).
    OFFLINE = 'offline'

    #: Corresponds to idle (yellow dot).
    IDLE = 'idle'

    #: Corresponds to Do Not Disturb (red dot).
    DND = 'dnd'

    #: Corresponds to invisible (gray dot).
    INVISIBLE = 'invisible'

    @property
    def strength(self) -> int:
        """
        :return: The strength of the presence, when computing the final presence on multiple \ 
            connections. 
        """
        return strengths.index(self)


strengths = [Status.OFFLINE, Status.INVISIBLE, Status.IDLE, Status.DND, Status.ONLINE]


class GameType(enum.IntEnum):
    """
    Represents a game's type.
    """
    #: Shows the ``Playing`` text.
    PLAYING = 0

    #: Shows the ``Streaming`` text.
    STREAMING = 1

    #: Shows the ``Listening to`` text.
    LISTENING_TO = 2

    #: Shows the ``Watching`` text.
    WATCHING = 3


class Game(object):
    """
    Represents a game object.
    """

    __slots__ = "type", "url", "name"

    def __init__(self, **kwargs) -> None:
        """
        :param name: The name for the game. 100 characters max.
        :param url: The URL for the game, if streaming.
        :param type: A :class:`.GameType` for this game.
        """
        #: The type of game this is.
        self.type = GameType.PLAYING  # type: GameType
        try:
            self.type = GameType(kwargs.get("type", 0))
        except ValueError:
            self.type = kwargs.get("type", 0)

        #: The stream URL this game is for.
        self.url = kwargs.get("url", None)  # type: str
        #: The name of the game being played.
        self.name = kwargs.get("name", None)  # type: str

    def to_dict(self) -> dict:
        """
        :return: The dict representation of this object. 
        """
        d = {
            "name": self.name,
            "type": self.type,
        }
        if self.url is not None:
            d["url"] = self.url

        return d

    def __repr__(self) -> str:
        return "<Game name='{}' type={} url={}>".format(self.name, self.type, self.url)


class Presence(object):
    """
    Represents a presence on a member.
    """

    __slots__ = "_status", "_game"

    def __init__(self, **kwargs) -> None:
        """
        :param status: The :class:`.Status` for this presence.
        :param game: The :class:`.Game` for this presence.
        """
        #: The :class:~.Status` for this presence.
        self._status = None  # type: Status
        # prevent dupe code by using our setter
        self.status = kwargs.get("status", Status.OFFLINE)

        game = kwargs.get("game", None)
        #: The :class:`.Game` for this presence.
        self._game = None  # type: Game

        # NB: this does a property set to ensure the types are right.
        self.game = game

    def __repr__(self):
        return "<Presence status={} game='{}'>".format(self.status, self.game)

    @property
    def status(self) -> Status:
        """
        :return: The :class:`.Status` associated with this presence.
        """
        return self._status

    @status.setter
    def status(self, value):
        if value is None:
            return

        if not isinstance(value, Status):
            value = Status(value)

        self._status = value

    @property
    def game(self) -> Game:
        """
        :return: The :class:`.Game` associated with this presence.
        """
        return self._game

    @game.setter
    def game(self, value):
        if value is None:
            self._game = None
            return

        if not isinstance(value, Game):
            value = Game(**value)

        self._game = value

    @property
    def strength(self) -> int:
        """
        :return: The strength for this status.
        """
        return self.status.strength


def _make_property(field: str, doc: str = None, max_size: int = None) -> property:
    def _getter(self):
        return self._rich_fields.get(field)

    def _setter(self, value: str):
        if max_size is not None and len(value) > max_size:
            raise ValueError("Field '{}' cannot be longer than {} characters"
                             .format(field, max_size))

        self._rich_fields[field] = value

    prop = property(_getter, _setter, doc=doc)
    return prop


class RichPresence(object):
    """
    Represents a Rich Presence. This class can be created safely for usage with :class:`.IPCClient`.
    """

    # typedef struct DiscordRichPresence {
    #     const char* state; /* max 128 bytes */
    #     const char* details; /* max 128 bytes */
    #     int64_t startTimestamp;
    #     int64_t endTimestamp;
    #     const char* largeImageKey; /* max 32 bytes */
    #     const char* largeImageText; /* max 128 bytes */
    #     const char* smallImageKey; /* max 32 bytes */
    #     const char* smallImageText; /* max 128 bytes */
    #     const char* partyId; /* max 128 bytes */
    #     int partySize;
    #     int partyMax;
    #     const char* matchSecret; /* max 128 bytes */
    #     const char* joinSecret; /* max 128 bytes */
    #     const char* spectateSecret; /* max 128 bytes */
    #     int8_t instance;
    # } DiscordRichPresence;
    def __init__(self, **fields):
        """
        :param fields: The rich presence fields.
        """
        self._rich_fields = fields

    state = _make_property("state", "The state for this presence.", 128)
    details = _make_property("details", "The details for this presence.", 128)

    @property
    def assets(self) -> dict:
        """
        The assets for this rich presence. Returns a dict of
        (large_image, large_text, small_image, small_text).
        """
        return self._rich_fields.get("assets", {})

    @assets.setter
    def assets(self, value: dict):
        for key in value.keys():
            if key not in ('large_image', 'large_text', 'small_image', 'small_text'):
                raise ValueError("Bad asset key: {}".format(key))

        self._rich_fields["assets"] = value

    @property
    def party_id(self) -> str:
        """
        The party ID for this rich presence.
        """
        return self._rich_fields.get("party", {}).get("id")

    @party_id.setter
    def party_id(self, value):
        if "party" not in self._rich_fields:
            self._rich_fields["party"] = {"id": value}
        else:
            self._rich_fields["party"]["id"] = value

    @property
    def party_size(self) -> List[int]:
        """
        The size of the party for this rich presence. An array of [size, max].
        """
        return self._rich_fields.get("party", {}).get("size")

    @party_size.setter
    def party_size(self, size: List[int]):
        if "party" not in self._rich_fields:
            self._rich_fields["party"] = {"size": size}
        else:
            self._rich_fields["party"]["size"] = size




# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Decorators to annotate function objects.

.. currentmodule:: curious.commands.decorators
"""
import inspect
import logging
from typing import Any, List, Type

from curious.commands import plugin as md_plugin
from curious.commands.ratelimit import BucketNamer, CommandRateLimit
from curious.commands.utils import get_description

logger = logging.getLogger(__name__)


def command(*,
            name: str = None, description: str = None,
            hidden: bool = False,
            aliases: List[str] = None, **kwargs):
    """
    Marks a function as a command. This annotates the command with some attributes that allow it
    to be invoked as a command.

    This decorator can be invoked like this:

    .. code-block:: python3

        @command()
        async def ping(self, ctx):
            await ctx.channel.messages.send("Ping!")

    :param name: The name of the command. If this is not specified, it will use the name of the \
        function object.
    :param description: The description of the command. If this is not specified, it will use the \
        first line of the docstring.
    :param hidden: If this command is hidden; i.e. it doesn't show up in the help listing.
    :param aliases: A list of aliases for this command.
    :param kwargs: Anything to annotate the command with.
    """

    # wrapper function that actually marks the object
    def inner(func):
        def set(attr: str, value: Any):
            try:
                return getattr(func, attr)
            except AttributeError:
                setattr(func, attr, value)

        set("is_cmd", True)
        set("cmd_name", name or func.__name__)
        set("cmd_description", description or get_description(func))
        set("cmd_aliases", aliases or [])
        set("cmd_subcommand", False)
        set("cmd_subcommands", [])
        set("cmd_parent", None)
        set("cmd_hidden", hidden)
        set("cmd_conditions", [])
        set("cmd_ratelimits", [])

        # annotate command object with any extra
        for ann_name, annotation in kwargs.items():
            ann_name = "cmd_" + name
            set(ann_name, annotation)

        func.subcommand = _subcommand(func)
        return func

    return inner


def condition(cbl):
    """
    Adds a condition to a command.

    This will add the callable to ``cmd_conditions`` on the function.
    """

    def inner(func):
        if not hasattr(func, "cmd_conditions"):
            func.cmd_conditions = []

        func.cmd_conditions.append(cbl)
        return func

    return inner


def ratelimit(*, limit: int, time: float, bucket_namer=BucketNamer.AUTHOR):
    """
    Adds a ratelimit to a command.
    """

    def inner(func):
        if not hasattr(func, "cmd_ratelimits"):
            func.cmd_ratelimits = []

        rl = CommandRateLimit(limit=limit, time=time, bucket_namer=bucket_namer)
        rl.command = func
        func.cmd_ratelimits.append(rl)
        return func

    return inner


def _subcommand(parent):
    """
    Decorator factory set on a command to produce subcommands.
    """

    def inner(**kwargs):
        # MULTIPLE LAYERS
        def inner_2(func):
            if not hasattr(parent, "is_cmd"):
                raise TypeError("Cannot be a subcommand of a non-command")

            cmd = command(**kwargs)(func)
            cmd.cmd_subcommand = True
            cmd.cmd_parent = parent
            parent.cmd_subcommands.append(cmd)
            return cmd

        return inner_2

    return inner


def autoplugin(plugin: 'Type[md_plugin.Plugin]' = None, *,
               startswith: str = "command") -> 'Type[md_plugin.Plugin]':
    """
    Automatically assigns commands inside a plugin.

    This will scan a :class:`.Plugin` for functions matching the pattern ``command_[parent_]name``,
    and automatically decorate them with the command decorator and subcommand decorators.

    :param plugin: The :class:`.Plugin` subclass to autoplugin.
    :param startswith: Used to override what the command function prefix will be.
    :return: The edited plugin.
    """
    # we were called like @autoplugin(startswith="something")
    # so return a lambda that provides the plugin
    if plugin is None:
        return lambda plugin: autoplugin(plugin, startswith=startswith)

    if not issubclass(plugin, md_plugin.Plugin):
        raise ValueError(f"Cannot autoplugin an object of type non-PluginMeta ({type(plugin)})")

    # we were called like @autoplugin or the lambda above completed
    logger.debug("Processing autocommand for plugin type %s", plugin.__name__)
    for name, member in plugin.__dict__.copy().items():
        if not name.startswith(startswith + "_"):
            continue

        # unwrap all functions
        member = inspect.unwrap(member)

        parts = name.split("_", 2)
        if len(parts) == 2:  # regular command, no parent
            # wrap it in a command and continue
            made_command = command(name=parts[1])(member)
            logger.debug("Made a top-level command %s on plugin %s", parts[1], plugin.__name__)
            setattr(plugin, name, made_command)
        else:
            # we have a parent, so call parent.subcommand() on it instead
            parent = parts[1]

            # o(n2) loop!
            def _pred(i):
                return hasattr(i, "cmd_name") and i.cmd_name == parent

            for _, found_command in inspect.getmembers(plugin, predicate=_pred):
                break
            else:
                raise AttributeError(
                    f"When doing an autoplugin, could not locate parent command {parent}.\n"
                    f"You need to make sure that the parent A) exists and B) is before the "
                    f"current command in the class definition for the autoplugin to resolve the "
                    f"name properly."
                ) from None

            made_command = found_command.subcommand(name=parts[2])(member)
            logger.debug("Made a subcommand %s (parent %s) on plugin %s",
                         parts[2], parts[1], plugin.__name__)
            setattr(plugin, name, made_command)

    return plugin



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Exceptions raised from within the library.

.. currentmodule:: curious.exc
"""
import enum
import warnings

from asks.response_objects import Response


class CuriousError(Exception):
    """
    The base class for all curious exceptions.
    """


# HTTP based exceptions.
class ErrorCode(enum.IntEnum):
    UNKNOWN_ACCOUNT = 10001
    UNKNOWN_APPLICATION = 10002
    UNKNOWN_CHANNEL = 10003
    UNKNOWN_GUILD = 10004
    UNKNOWN_INTEGRATION = 10005
    UNKNOWN_INVITE = 10006
    UNKNOWN_MEMBER = 10007
    UNKNOWN_MESSAGE = 10008
    UNKNOWN_OVERWRITE = 1009
    UNKNOWN_PROVIDER = 10010
    UNKNOWN_ROLE = 10011
    UNKNOWN_TOKEN = 10012
    UNKNOWN_USER = 10013
    UNKNOWN_EMOJI = 10014

    NO_BOTS = 20001
    ONLY_BOTS = 20002

    MAX_GUILDS = 30001  # technically user only
    MAX_FRIENDS = 30002
    MAX_PINS = 30003
    MAX_ROLES = 30005
    MAX_REACTIONS = 30010
    MAX_GUILD_CHANNELS = 30013

    UNAUTHORIZED = 40001
    MISSING_ACCESS = 50001
    INVALID_ACCOUNT = 50002
    NO_DMS = 50003
    EMBED_DISABLED = 50004
    CANNOT_EDIT = 50005
    CANNOT_SEND_EMPTY_MESSAGE = 50006
    CANNOT_SEND_TO_USER = 50007
    CANNOT_SEND_TO_VC = 50008
    VERIFICATION_TOO_HIGH = 50009

    OAUTH2_NO_BOT = 50010
    OAUTH2_LIMIT = 50011
    INVALID_OAUTH_STATE = 50012
    MISSING_PERMISSIONS = 50013
    INVALID_AUTH_TOKEN = 50014

    NOTE_TOO_LONG = 50015
    INVALID_MESSAGE_COUNT = 50016
    CANNOT_PIN = 50019
    INVALID_VANITY_URL = 50020
    TOO_OLD_TO_BULK_DELETE = 50034
    INVALID_FORM_BODY = 50035
    INVALID_INVITE_GUILD = 50036

    REACTION_BLOCKED = 90001

    UNKNOWN = 0


class HTTPException(CuriousError, ConnectionError):
    """
    Raised when a HTTP request fails with a 400 <= e < 600 error code.
    """

    def __init__(self, response: Response, error: dict):
        self.response = response

        error_code = error.get("code", 0)
        try:
            #: The error code for this response.
            self.error_code = ErrorCode(error_code)
        except ValueError:
            warnings.warn("Received unknown error code {}")
            #: The error code for this response.
            self.error_code = ErrorCode.UNKNOWN
        self.error_message = error.get("message")

        self.error = error

    def __str__(self) -> str:
        if self.error_code == ErrorCode.UNKNOWN:
            return repr(self.error)

        return "{} ({}): {}".format(self.error_code, self.error_code.name, self.error_message)

    __repr__ = __str__


class Unauthorized(HTTPException):
    """
    Raised when your bot token is invalid.
    """


class Forbidden(HTTPException):
    """
    Raised when you don't have permission for something.
    """


class NotFound(HTTPException):
    """
    Raised when something could not be found.
    """


class PermissionsError(CuriousError, PermissionError):
    """
    Raised when you do not have sufficient permission to perform an action.

    :ivar permission_required: The string of the permission required to perform this action.
    """

    def __init__(self, permission_required: str):
        self.permission_required = permission_required

    def __str__(self) -> str:
        return "Bot requires the permission {} to perform this action"\
            .format(self.permission_required)

    __repr__ = __str__


class HierarchyError(CuriousError, PermissionError):
    """
    Raised when you can't do something due to the hierarchy.
    """



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Defines commands-specific exceptions.

.. currentmodule:: curious.commands.exc
"""
import time
from math import ceil
from typing import Tuple

from curious.exc import CuriousError


class CommandsError(CuriousError):
    pass


class ConditionsFailedError(CommandsError):
    """
    Raised when conditions fail for a command.
    """
    def __init__(self, ctx, check):
        self.ctx = ctx
        self.conditions = check

    def __repr__(self) -> str:
        if isinstance(self.conditions, list):
            return f"The conditions for `{self.ctx.command_name}` failed."

        return f"The condition `{self.conditions.__name__}` for `{self.ctx.command_name}` failed." \

    __str__ = __repr__


class MissingArgumentError(CommandsError):
    """
    Raised when a command is missing an argument.
    """
    def __init__(self, ctx, arg):
        self.ctx = ctx
        self.arg = arg

    def __repr__(self) -> str:
        return f"Missing required argument `{self.arg}` in `{self.ctx.command_name}`."

    __str__ = __repr__


class CommandInvokeError(CommandsError):
    """
    Raised when a command has an error during invokation.
    """
    def __init__(self, ctx):
        self.ctx = ctx

    def __repr__(self) -> str:
        return f"Command {self.ctx.command_name} failed to invoke with error `{self.__cause__}`."

    __str__ = __repr__


class ConversionFailedError(CommandsError):
    """
    Raised when conversion fails.
    """
    def __init__(self, ctx, arg: str, to_type: type, message: str = "Unknown error"):
        self.ctx = ctx
        self.arg = arg
        self.to_type = to_type
        self.message = message

    def __repr__(self) -> str:
        return f"Cannot convert `{self.arg}` to type `{self.to_type.__name__}`: {self.message}."

    __str__ = __repr__


class CommandRateLimited(CommandsError):
    """
    Raised when a command is ratelimited.
    """
    def __init__(self, context, func, limit, bucket: Tuple[int, float]):
        self.ctx = context
        self.func = func
        self.limit = limit
        self.bucket = bucket

    def __repr__(self) -> str:
        left = int(ceil(self.bucket[1] - time.monotonic()))
        return f"The command {self.ctx.command_name} is currently rate limited for " \
               f"{left} second(s)."

    __str__ = __repr__


# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Search objects.

.. currentmodule:: curious.dataclasses.search
"""
import collections
import functools
import typing
from typing import Iterator

from curious.dataclasses import channel as dt_channel, guild as dt_guild, member as dt_member, \
    message as dt_message, user as dt_user


class MessageGroup:
    """
    A small class that returns messages from a message group.
    """

    __slots__ = "msgs",

    def __init__(self, msgs: 'typing.List[dt_message.Message]'):
        self.msgs = msgs

    # generic magic methods
    def __getitem__(self, item) -> 'dt_message.Message':
        return self.msgs[item]

    def __iter__(self) -> 'Iterator[dt_message.Message]':
        return iter(self.msgs)

    def __repr__(self) -> str:
        return "<MessageGroup msgs='{}'>".format(self.msgs)

    @property
    def before(self) -> 'typing.Tuple[dt_message.Message, dt_message.Message]':
        """
        :return: The two :class:`.Message` objects that happen before the requested message.
        """
        return self.msgs[0], self.msgs[1]

    @property
    def message(self) -> 'dt_message.Message':
        """
        :return: The :class:`.Message` that matched this search query.
        """
        return self.msgs[2]

    @property
    def after(self) -> 'typing.Tuple[dt_message.Message, dt_message.Message]':
        """
        :return: The two :class:`.Message` objects that happen after the requested message.
        """
        return self.msgs[3], self.msgs[4]


class SearchResults(collections.AsyncIterator):
    """
    An async iterator that can be used to iterate over the results of a search.
    This will automatically fill results, and return messages as appropriate.
    
    The return type of iterating over this is a :class:`.MessageGroup`, which contains the messages
    around the message that matched the search result.
    
    .. code-block:: python3
    
        async for i in sr:
            print(i.before)  # 2 messages from before
            print(i.message) # the message that matched
            print(i.after)   # 2 messages from after
            
    """
    def __init__(self, sq: 'SearchQuery') -> None:
        self.sq = sq

        # state vars
        self.page = 0
        self.groups = collections.deque()

        self._limit = -1
        self._total_count = 0

    def __repr__(self) -> str:
        return "<SearchResults page='{}' messages='{}'>".format(self.page, len(self.groups))

    # builder methods
    def limit(self, limit: int=-1) -> 'SearchResults':
        """
        Sets the maximum messages to fetch from this search result.
        
        .. code-block:: python3
        
            async for group in sr.limit(25):
                ...
        
        :param limit: The limit to set.
        :return: This :class:`.SearchResults`.
        """
        self._limit = limit
        return self

    async def fetch_next_page(self) -> None:
        """
        Fetches the next page of results from the SearchQuery.
        """
        if self._limit != -1 and self._total_count >= self._limit:
            return

        results = await self.sq.execute(page=self.page)

        # add a new messagegroup to the end
        for r in results:
            self.groups.append(MessageGroup(r))

        self.page += 1

    def get_next(self) -> 'MessageGroup':
        """
        Gets the next page of results.
        
        If no results were found, this will raise an IndexError, and you must fetch the next page 
        with :meth:`.SearchResults.fetch_next_page`.
        
        :return: A :class:`.MessageGroup` for the next page of results, if applicable.
        """
        # prevent more fetching
        if self._limit != -1 and self._total_count >= self._limit:
            raise IndexError

        popped = self.groups.popleft()
        self._total_count += len(popped.msgs)
        return popped

    async def __anext__(self) -> 'MessageGroup':
        try:
            return self.get_next()
        except IndexError:
            await self.fetch_next_page()
            # try and pop left again
            # if it fails no messages were returned
            try:
                return self.get_next()
            except IndexError:
                raise StopAsyncIteration


class SearchQuery(object):
    """
    Represents a search query to be sent to Discord. This is a simple wrapper over the HTTP API.
    
    For example, to search a channel called ``general`` for messages with the content ``heck``:
    
    .. code-block:: python3

        with ctx.guild.search as sq:
            sq.content = "heck"
            sq.channel = next(filter(lambda c: c.name == "general", ctx.guild.channels), None)
             
        async for result in sq.results:
            ...  # do whatever with the messages returned.
            
    You can get results out of the query in two ways:
    
    .. code-block:: python3
    
        sq = SearchQuery(ctx.guild)
        sq.content = "heck"
        
        # form 1
        async for item in sq.results:
            ...
            
        # form 2
        results = await sq.get_messages()
        for result in results:
            ...
            
    It is recommended to use the ``async for`` form, as this will automatically page the results 
    and return the next page of results as soon as the current one is exhausted.
    """

    def __init__(self, guild: 'dt_guild.Guild' = None,
                 channel: 'dt_channel.Channel' = None) -> None:
        """
        :param guild: The :class:`.Guild` to search the messages for.
        :param channel: The :class:`.Channel` to search messages for. Only used for DMs.
        """
        self._guild = guild

        # internal vars used for the search
        self._channel = channel
        self._query = None  # type: str
        self._author = None  # type: typing.Union[dt_user.User, dt_member.Member]

    def make_params(self) -> typing.Dict[str, str]:
        """
        :return: The dict of parameters to send for this request. 
        """
        params = {}

        if self.guild is not None and self.channel is not None:
            params["channel_id"] = self.channel.id

        if self._query is not None:
            params["content"] = self._query

        if self._author is not None:
            params["author_id"] = self._author.id

        # TODO: Datetimes and `has:`

        return params

    # magic methods
    def __enter__(self) -> 'SearchQuery':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        return False

    def __repr__(self) -> str:
        return "<SearchQuery guild='{}' channel='{}'>".format(self.guild, self.channel)

    # internal properties
    @property
    def _http_meth(self) -> typing.Callable[[], dict]:
        """
        :return: The built URL to execute this search query on. 
        """
        if self.guild is not None:
            return functools.partial(self._bot.http.search_guild, self.guild.id)

        return functools.partial(self._bot.http.search_channel, self.channel.id)

    @property
    def _bot(self):
        if self._guild is not None:
            return self._guild._bot

        return self._channel._bot

    # public properties
    @property
    def guild(self) -> 'typing.Union[dt_guild.Guild, None]':
        """
        :return: The :class:`.Guild` this search query is searching.
        """
        return self._guild

    @property
    def channel(self) -> 'typing.Union[dt_channel.Channel, None]':
        """
        The :class:`.Channel` that is being searched.
        
        .. note::
            
            If this a DM, this will not be added in the params.
        
        :getter: Gets the :class:`.Channel` to be searched.
        :setter: Sets the :class:`.Channel` to be searched.
        """
        return self._channel

    @channel.setter
    def channel(self, value):
        if not isinstance(value, dt_channel.Channel):
            raise TypeError("Must provide a Channel object")

        if value.type is dt_channel.ChannelType.VOICE:
            raise ValueError("Cannot search a voice channel")

        if self._guild is not None and value.guild is None:
            raise ValueError("Channel must not be a private channel for searching a guild")

        if self._guild is not None and value.guild != self._guild:
            raise ValueError("Channel to search must be in the same guild")

        self._channel = value

    @property
    def content(self) -> str:
        """
        The str content that is being searched.
        
        :getter: Gets the ``str`` content to be searched.
        :setter: Sets the ``str`` content to be searched.
        """
        return self._query

    @content.setter
    def content(self, value):
        self._query = value

    @property
    def results(self) -> 'SearchResults':
        """
        A simple way of accessing the search results for a search query.
        
        :return: A :class:`.SearchResults` representing the results of this query.
        """
        return SearchResults(self)

    # workhouse methods
    async def execute(self, page: int = 0) -> 'typing.List[typing.List[dt_message.Message]]':
        """
        Executes the search query.
        
        .. warning::
            
            This is an internal method, used by the library. Use :meth:`.get_messages` instead
            of this.
        
        :param page: The page of results to return.
        :return: A list of :class:`.Message` which returns the results of the search query.
        """
        func = self._http_meth
        params = self.make_params()

        # get the offset page
        params["offset"] = page * 25
        # make the http request
        res = await func(params)

        message_blocks = []

        # parse all of the message objects
        for group in res.get("messages", []):
            message_blocks.append([self._bot.state.make_message(m) for m in group])

        return message_blocks

    async def get_messages(self, page: int = 0) -> 'SearchResults':
        """
        Executes the search query and gets the messages for the specified page.
        
        :param page: The page of results to return. 
        :return: A :class:`.SearchResult` that can be used to search the results.
        """
        res = SearchResults(self)
        res.page = page
        await res.fetch_next_page()
        return res



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Widget objects.

.. currentmodule:: curious.dataclasses.widget
"""
import typing
from types import MappingProxyType

from curious.dataclasses import channel as dt_channel, guild as dt_guild
from curious.dataclasses.bases import Dataclass
from curious.dataclasses.presence import Game, Status


class WidgetChannel(Dataclass):
    """
    Represents a limited subsection of a channel.
    """

    def __init__(self, bot, guild: 'WidgetGuild', **kwargs):
        super().__init__(id=int(kwargs.get("id", 0)), cl=bot)

        #: The name of this channel.
        self.name = kwargs.get("name")

        #: The position of this channel.
        self.position = kwargs.get("position", -1)

        #: The guild ID for this channel.
        self.guild_id = guild.id

        #: The :class:`.WidgetGuild` for this channel.
        self.guild = guild


class WidgetMember(Dataclass):
    """
    Represents a limited subsection of a member.
    """

    def __init__(self, bot, guild: 'WidgetGuild', kwargs):
        super().__init__(id=int(kwargs.get("id", 0)), cl=bot)

        # construct a superficial user dict
        user_dict = {
            "id": self.id,
            "name": kwargs.get("name", None),
            "avatar": kwargs.get("avatar", None),
            "discriminator": kwargs.get("discriminator", None),
            "bot": kwargs.get("bot", False)
        }
        #: The :class:`.User` object associated with this member.
        self.user = bot.state.make_user(user_dict)
        bot.state._check_decache_user(user_dict["id"])

        #: The :class:`.WidgetGuild` object associated with this member.
        self.guild = guild

        #: The game associated with this member.
        game = kwargs.get("game")
        if game is None:
            game = {}
        self.game = Game(**game) if game else None

        #: The :class:`.Status` associated with this member.
        self.status = Status(kwargs.get("status"))


class WidgetGuild(Dataclass):
    """
    Represents a limited subsection of a guild.
    """

    def __init__(self, bot, **kwargs):
        super().__init__(id=int(kwargs.get("id", 0)), cl=bot)

        #: The name of this guild.
        self.name = kwargs.get("name", "")

        #: A mapping of :class:`.WidgetChannel` in this widget guild.
        self._channels = {}  # type: typing.MutableMapping[int, WidgetChannel]
        for channel in kwargs.get("channels", []):
            c = WidgetChannel(bot=self._bot, guild=self, **channel)
            self._channels[c.id] = c

        #: A mapping of :class:`.WidgetMember` in this widget guild.
        self._members = {}
        for member in kwargs.get("members", []):
            m = WidgetMember(bot=self._bot, guild=self, kwargs=member)
            self._members[m.id] = m

    @property
    def channels(self) -> 'typing.Mapping[int, WidgetChannel]':
        """
        :return: A read-only mapping of :class:`.WidgetChannel` representing the channels for \
            this guild. 
        """
        return MappingProxyType(self._channels)

    @property
    def members(self) -> 'typing.Mapping[int, WidgetMember]':
        """
        :return: A read-only mapping of :class:`.WidgetMember` representing the channels for \
            this guild. 
        """
        return MappingProxyType(self._members)

    def __repr__(self):
        return "<WidgetGuild id={} members={} name='{}'>".format(self.id, len(self.members),
                                                                 self.name)

    __str__ = __repr__


class Widget(object):
    """
    Represents the embed widget for a guild.
    """

    def __init__(self, client, **kwargs):
        self._bot = client

        #: The guild ID for this widget.
        self.guild_id = int(kwargs.get("id", 0))

        #: The widget guild for this widget.
        self._widget_guild = WidgetGuild(self._bot, **kwargs)

        #: The invite URL that this widget represents.
        self.invite_url = kwargs.get("instant_invite", None)

    @property
    def guild(self) -> 'typing.Union[dt_guild.Guild, WidgetGuild]':
        """
        :return: The guild object associated with this widget.
            If the guild was cached, a :class:`.Guild`. Otherwise, a :class:`.WidgetGuild`.
        """
        try:
            return self._bot.guilds[self.guild_id]
        except KeyError:
            return self._widget_guild

    @property
    def channels(self) -> 'typing.Mapping[int, typing.Union[dt_channel.Channel, WidgetChannel]]':
        """
        :return: A mapping of channels associated with this widget.
        """
        return self.guild.channels

    def __repr__(self):
        return "<Widget guild={}>".format(self.guild)
